import unittest
import warnings
import pandas as pd
import numpy as np
import pathlib
from testfixtures import TempDirectory
import gc

import tensorflow.keras.backend as K

from ml4ir.applications.ranking.pipeline import RankingPipeline
from ml4ir.applications.ranking.config.parse_args import get_args

warnings.filterwarnings("ignore")


def ml4ir_sanity_pipeline(df, working_dir, log_dir, n_features):
    """
    Train ml4ir on the passed data and read the MRR for ml4ir
    """
    df.to_csv(working_dir / 'train' / 'data.csv')
    df.to_csv(working_dir / 'validation' / 'data.csv')
    df.to_csv(working_dir / 'test' / 'data.csv')

    fconfig_name = "feature_config_sanity_tests_" + str(n_features) + "_features.yaml"
    feature_config_file = pathlib.Path(__file__).parent / "data" / "configs" / fconfig_name
    model_config_file = pathlib.Path(__file__).parent / "data" / "configs" / "model_config_sanity_tests.yaml"
    train_ml4ir(working_dir.as_posix(), feature_config_file.as_posix(), model_config_file.as_posix(), log_dir.as_posix())
    ml4ir_results = pd.read_csv(log_dir / 'test_command_line'/ '_SUCCESS', header=None)
    ml4ir_mrr = float(ml4ir_results.loc[ml4ir_results[0] == 'test_new_MRR'][1])
    return ml4ir_mrr


def train_ml4ir(data_dir, feature_config, model_config, logs_dir):
    """
    Train a pointwise ranker, listwise loss model using ml4ir
    """
    argv = ["--data_dir", data_dir,
            "--feature_config", feature_config,
            "--loss_type", "listwise",
            "--scoring_type", "listwise",
            "--run_id", "test_command_line",
            "--data_format", "csv",
            "--execution_mode", "train_evaluate",
            "--loss_key", "rank_one_listnet",
            "--num_epochs", "150",
            "--model_config", model_config,
            "--batch_size", "1",
            "--logs_dir", logs_dir,
            "--max_sequence_size", "25",
            "--train_pcent_split", "1.0",
            "--val_pcent_split", "-1",
            "--test_pcent_split", "-1",
            "--early_stopping_patience", "25",
            "--metrics_keys", "MRR", "categorical_accuracy",
            "--monitor_metric", "categorical_accuracy"]
    args = get_args(argv)
    rp = RankingPipeline(args=args)
    rp.run()


def run_sanity_test(n_features, fname, perceptron_mrr, log_regression_mrr, working_dir, log_dir):
    """
    Runs sanity test for linear models.
    Parameters
    ----------
    n_features : int
        Number of features in the dataset.
    fname : string
        Dataset file name
    perceptron_mrr : float
        MRR of perceptron
    log_regression_mrr : float
        MRR of logistic regression
    working_dir : string
        Path of the working directory
    log_dir : string
        Path of the log directory
    """
    df = pd.read_csv(pathlib.Path(__file__).parent / "data" / "linear_sanity_tests" / fname)
    ml4ir_mrr = ml4ir_sanity_pipeline(df, working_dir, log_dir, n_features)

    assert np.isclose(ml4ir_mrr, perceptron_mrr, atol=0.001) or ml4ir_mrr >= perceptron_mrr
    assert np.isclose(ml4ir_mrr, log_regression_mrr, atol=0.001) or ml4ir_mrr >= log_regression_mrr


class TestML4IRSanity(unittest.TestCase):
    """
    Datasets used in these tests are generated by sklearn.datasets.make_classification by changing its parameters.
    Then queries and documents are grouped randomly (without replacement) so that each query has 2-25 documents with
    one clicked document.
    Baseline models (Perceptron and logistic regression are trained with sklearn)
    """

    def setUp(self):
        self.dir = pathlib.Path(__file__).parent
        self.working_dir = TempDirectory()
        self.log_dir = self.working_dir.makedir('logs')
        self.working_dir.makedir('train')
        self.working_dir.makedir('test')
        self.working_dir.makedir('validation')

    def tearDown(self):
        TempDirectory.cleanup_all()

        # Explicitly clear keras memory
        gc.collect()
        K.clear_session()

    def test_linear_ml4ir_sanity_1(self):
        """
        Weights of pre-trained models:
            perceptron=np.array([1.87212065, -0.00305068])
            log_regression=np.array([28.62071696, 1.18915853])
        """
        run_sanity_test(n_features=2, fname="dataset1.csv",
                        perceptron_mrr=1.0,
                        log_regression_mrr=1.0,
                        working_dir=pathlib.Path(self.working_dir.path), log_dir=pathlib.Path(self.log_dir))

    def test_linear_ml4ir_sanity_2(self):
        """
        Weights of pre-trained models:
            perceptron=np.array([4.50209484, -0.80280452])
            log_reg=np.array([22.73585585, -3.94821153])
        """
        run_sanity_test(n_features=2, fname="dataset2.csv",
                        perceptron_mrr=1.0,
                        log_regression_mrr=1.0,
                        working_dir=pathlib.Path(self.working_dir.path), log_dir=pathlib.Path(self.log_dir))

    def test_linear_ml4ir_sanity_3(self):
        """
        Weights of pre-trained models:
            perceptron=np.array([-1.27651475, -4.07647092, 8.23950305, 0.29241316, 3.24763417])
            log_reg=np.array([-1.67270377, -5.76088727, 8.36278576, -0.90878154, 3.47653204])
        """
        run_sanity_test(n_features=5, fname="dataset3.csv",
                        perceptron_mrr=1.0,
                        log_regression_mrr=1.0,
                        working_dir=pathlib.Path(self.working_dir.path), log_dir=pathlib.Path(self.log_dir))

    def test_linear_ml4ir_sanity_4(self):
        """
        Weights of pre-trained models:
            perceptron=np.array([5.10535665, 1.44131417])
            log_reg=np.array([20.0954756, 4.69360163])
        """
        run_sanity_test(n_features=2, fname="dataset4.csv",
                        perceptron_mrr=0.975609756097561,
                        log_regression_mrr=0.9878048780487805,
                        working_dir=pathlib.Path(self.working_dir.path), log_dir=pathlib.Path(self.log_dir))

    def test_linear_ml4ir_sanity_5(self):
        """
        Weights of pre-trained models:
            perceptron=np.array([0.57435291, -0.99437351])
            log_reg=np.array([1.15593505, -0.93317691])
        """
        run_sanity_test(n_features=2, fname="dataset5.csv",
                        perceptron_mrr=0.37021768881524975,
                        log_regression_mrr=0.4290204375570229,
                        working_dir=pathlib.Path(self.working_dir.path), log_dir=pathlib.Path(self.log_dir))

    def test_linear_ml4ir_sanity_6(self):
        """
        Weights of pre-trained models:
        perceptron=np.array(
                        [4.59994733, -3.56373965, -6.15935686, 0.87523846, -0.64231058, 2.15971991, 5.79875003,
                         -7.70152594, -0.07521741, 2.8817456])
        log_reg=np.array(
                        [-0.38064406, -0.27970534, 0.02775136, 0.25641926, 0.15413321, 0.29194965, 0.72707686,
                         0.24791729, -0.39367192, 0.4882174])
        """
        run_sanity_test(n_features=10, fname="dataset6.csv",
                        perceptron_mrr=0.2630707011921534,
                        log_regression_mrr=0.29970690874564615,
                        working_dir=pathlib.Path(self.working_dir.path), log_dir=pathlib.Path(self.log_dir))

    def test_linear_ml4ir_sanity_7(self):
        """
        Weights of pre-trained models:
            perceptron=np.array([0.40127356, -0.43773627])
            log_reg=np.array([4.15630544, -1.09111369])
        """
        run_sanity_test(n_features=2, fname="dataset7.csv",
                        perceptron_mrr=0.304679599925838,
                        log_regression_mrr=0.4938108725899423,
                        working_dir=pathlib.Path(self.working_dir.path), log_dir=pathlib.Path(self.log_dir))

    def test_linear_ml4ir_sanity_8(self):
        """
        Weights of pre-trained models:
            perceptron=np.array(
                            [2.91798129, 4.24880336, 7.42919018, 2.49609694, -0.84988373, 0.43435823, -0.18953416,
                             2.23129287, -0.67951411, -0.63925108])
            log_reg=np.array(
                            [-0.14472192, -0.22594271, 0.62703883, 0.16002515, 0.17084088, -0.22872226, 0.89200279,
                             0.06297475, 0.70470567, -0.19396659])
        """
        run_sanity_test(n_features=10, fname="dataset8.csv",
                        perceptron_mrr=0.31185341006782435,
                        log_regression_mrr=0.3092000998257748,
                        working_dir=pathlib.Path(self.working_dir.path), log_dir=pathlib.Path(self.log_dir))


if __name__ == "__main__":
    unittest.main()



from tensorflow.keras import Input
import tensorflow as tf

from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.features.feature_layer import FeatureLayerMap
from ml4ir.base.features.feature_layer import define_feature_layer
from ml4ir.base.io.file_io import FileIO

from typing import Dict


class InteractionModel:
    """
    InteractionModel class that defines tensorflow layers that act on input features to
    convert them into numeric features to be fed into further neural network layers
    """

    def __call__(self, inputs: Dict[str, Input]):
        """
        Convert input tensorflow features into numeric
        train features and metadata features by applying respective
        feature transformation functions as specified in the FeatureConfig

        Parameters
        ----------
        inputs : dict
            Dictionary of the inputs to the tensorflow keras model

        Returns
        -------
        train_features : `tf.Tensor`
            Dense tensor object that is used for training
        metadata_features : dict
            Dictionary of feature tensors that can be used for
            computing custom metrics and losses
        """
        train_features, metadata_features = self.feature_layer_op(inputs)
        train_features, metadata_features = self.transform_features_op(
            train_features, metadata_features
        )
        return train_features, metadata_features

    def feature_layer_op(self, inputs: Dict[str, Input]):
        """
        Convert input tensorflow features into numeric
        train features and metadata features by applying respective feature
        transformation functions as specified in the FeatureConfig

        Parameters
        ----------
        inputs : dict
            Dictionary of the inputs to the tensorflow keras model

        Returns
        -------
        train_features : dict
            Dict of feature tensors that are used for training
        metadata_features : dict
            Dictionary of feature tensors that can be used for
            computing custom metrics and losses
        """
        raise NotImplementedError

    def transform_features_op(self, train_features, metadata_features):
        """
        Convert train and metadata features which have feature layer
        functions applied to them into dense numeric tensors

        Parameters
        ----------
        inputs : dict
            Dictionary of the inputs to the tensorflow keras model

        Returns
        -------
        train_features : `tf.Tensor`
            Dense tensor object that is used for training
        metadata_features : dict
            Dictionary of feature tensors that can be used for
            computing custom metrics and losses
        """
        raise NotImplementedError


class UnivariateInteractionModel(InteractionModel):
    """
    Defines an interaction model that configures feature layer operations
    on individual features
    """

    def __init__(
        self,
        feature_config: FeatureConfig,
        tfrecord_type: str,
        feature_layer_keys_to_fns: dict = {},
        max_sequence_size: int = 0,
        file_io: FileIO = None,
    ):
        """
        Constructor for instantiating a UnivariateInteractionModel

        Parameters
        ----------
        feature_config : `FeatureConfig` object
            FeatureConfig object that defines list of model features
            and the feature transformation functions to be used on each
        tfrecord_type : {"example", "sequence_example"}
            Type of TFRecord protobuf being used for model training
        feature_layer_keys_to_fns : dict
            Dictionary of custom feature transformation functions to be applied
            on the input features
        max_sequence_size : int, optional
            Maximum size of the sequence in SequenceExample protobuf
        file_io : FileIO object
            `FileIO` object that handles read write operations
        """
        self.feature_config = feature_config
        self.tfrecord_type = tfrecord_type
        self.max_sequence_size = max_sequence_size
        self.feature_layer_map = FeatureLayerMap()
        self.feature_layer_map.add_fns(feature_layer_keys_to_fns)
        self.file_io = file_io

    def feature_layer_op(self, inputs: Dict[str, Input]):
        """
        Convert input tensorflow features into numeric
        train features and metadata features by applying respective feature
        transformation functions as specified in the FeatureConfig

        Parameters
        ----------
        inputs : dict
            Dictionary of the inputs to the tensorflow keras model

        Returns
        -------
        train_features : dict
            Dict of feature tensors that are used for training
        metadata_features : dict
            Dictionary of feature tensors that can be used for
            computing custom metrics and losses
        """
        train_features, metadata_features = define_feature_layer(
            feature_config=self.feature_config,
            tfrecord_type=self.tfrecord_type,
            feature_layer_map=self.feature_layer_map,
            file_io=self.file_io,
        )(inputs)

        return train_features, metadata_features

    def transform_features_op(
        self, train_features: Dict[str, tf.Tensor], metadata_features: Dict[str, tf.Tensor]
    ):
        """
        Convert train and metadata features which have feature layer
        functions applied to them into dense numeric tensors.
        Sorts the features by name and concats the individual features
        into a dense tensor.

        Parameters
        ----------
        inputs : dict
            Dictionary of the inputs to the tensorflow keras model

        Returns
        -------
        train_features : `tf.Tensor`
            Dense tensor object that is used for training
        metadata_features : dict
            Dictionary of feature tensors that can be used for
            computing custom metrics and losses
        """
        # Sorting the train features dictionary so that we control the order
        train_features_list = [train_features[k] for k in sorted(train_features)]

        # Concat all train features to get a dense feature vector
        train_features_transformed = tf.concat(train_features_list, axis=-1, name="train_features")

        return train_features_transformed, metadata_features



import pandas as pd
import numpy as np

from typing import List, Union

# Metrics where higher value is better/desirable
POSITIVE_METRIC_SUFFIXES = [
    "MRR",
    "NDCG_mean"
]
# Metrics where lower value is better/desirable
NEGATIVE_METRIC_SUFFIXES = [
    "ACR",
    "failure_all_mean",
    "failure_any_mean",
    "failure_all_rank_mean",
    "failure_any_rank_mean",
    "failure_any_count_mean",
    "failure_any_fraction_mean"
]


def compute_dcg(relevance_grades: List[float]):
    """
    Compute the discounted cumulative gains on a list of relevance grades

    Parameters
    ----------
    relevance_grades: list of float
        Relevance grades to be used to compute DCG metric
        The rank is the position in the list

    Returns
    -------
    float
        Computed DCG for the ranked list of relevance grades

    Notes
    -----
    Reference -> https://en.wikipedia.org/wiki/Discounted_cumulative_gain
    """
    return np.sum([(np.power(2, relevance_grades[i] - 1.) / np.log2(i + 1 + 1)) for i in range(len(relevance_grades))])


def compute_ndcg(relevance_grades: List[float]):
    """
    Compute the normalized discounted cumulative gains on a list of relevance grades

    Parameters
    ----------
    relevance_grades: list of float
        Relevance grades to be used to compute NDCG metric
        The rank is the position in the list

    Returns
    -------
    float
        Computed NDCG for the ranked list of relevance grades

    Notes
    -----
    Reference -> https://en.wikipedia.org/wiki/Discounted_cumulative_gain
    """
    return compute_dcg(relevance_grades) / compute_dcg(sorted(relevance_grades, reverse=True))


def compute_secondary_label_metrics(
    secondary_label_values: pd.Series,
    ranks: pd.Series,
    click_rank: int,
    secondary_label: str,
    prefix: str = "",
):
    """
    Computes the secondary ranking metrics using a secondary label for a single query

    Parameters
    ----------
    secondary_label_values: pd.Series
        Series object containing the secondary label values for a given query
    ranks: pd.Series
        Series object containing the ranks corresponding to the secondary label values
    click_rank: int
        Rank of the clicked record
    secondary_label: str
        Name of the secondary label used to define metric name
    prefix: str
        Prefix attached to the metric name

    Returns
    -------
    dict
        Key value pairs of the metric names and the associated computed values
        using the secondary label

    Notes
    -----
    A secondary label is any feature/value that serves as a proxy relevance assessment that
    the user might be interested to measure on the dataset in addition to the primary click labels.
    For example, this could be used with an exact query match feature. In that case, the metric
    sheds light on scenarios where the records with an exact match are ranked lower than those without.
    This would provide the user with complimentary information (to typical click metrics such as MRR and ACR)
    about the model to help make better trade-off decisions w.r.t. best model selection.
    """
    failure_all = 0
    failure_any = 0
    failure_count = 0
    failure_fraction = 0.0

    try:
        click_secondary_label_value = secondary_label_values[ranks == click_rank].values[0]
        pre_click_secondary_label_values = secondary_label_values[ranks < click_rank]

        if pre_click_secondary_label_values.size > 0:
            # Query failure only if failure on all records
            failure_all = (
                1
                if (pre_click_secondary_label_values < click_secondary_label_value).all()
                else 0
            )
            # Query failure if failure on at least one record
            failure_any = (
                1
                if (pre_click_secondary_label_values < click_secondary_label_value).any()
                else 0
            )
            # Count of failure records
            failure_count = (
                pre_click_secondary_label_values < click_secondary_label_value
            ).sum()
            # Normalizing to fraction of potential records
            failure_fraction = failure_count / (click_rank - 1)

    except IndexError:
        # Ignore queries with missing or invalid click labels
        pass

    # Compute NDCG metric on the secondary label
    # NOTE: Here we are passing the relevance grades ordered by the ranking
    secondary_label_ndcg = compute_ndcg(
        secondary_label_values.values[np.argsort(ranks.values)])

    return {
        "{}{}_NDCG".format(prefix, secondary_label): secondary_label_ndcg,
        "{}{}_failure_all".format(prefix, secondary_label): failure_all,
        "{}{}_failure_any".format(prefix, secondary_label): failure_any,
        "{}{}_failure_all_rank".format(prefix, secondary_label): click_rank
        if failure_all
        else 0,
        "{}{}_failure_any_rank".format(prefix, secondary_label): click_rank
        if failure_any
        else 0,
        "{}{}_failure_any_count".format(prefix, secondary_label): failure_count,
        "{}{}_failure_any_fraction".format(prefix, secondary_label): failure_fraction,
    }


def compute_secondary_labels_metrics_on_query_group(
    query_group: pd.DataFrame,
    label_col: str,
    old_rank_col: str,
    new_rank_col: str,
    secondary_labels: List[str],
    group_keys: List[str] = []
):
    """
    Compute the old and new secondary ranking metrics for a given
    query on a list of secondary labels

    Parameters
    ----------
    query_group : `pd.DataFrame` object
        DataFrame group object for a single query to compute secondary metrics on
    label_col : str
        Name of the label column in the query_group
    old_rank_col : str
        Name of the column that represents the original rank of the records
    new_rank_col : str
        Name of the column that represents the newly computed rank of the records
        after reordering based on new model scores
    secondary_labels : list
        List of features used to compute secondary metrics
    group_keys : list, optional
        List of features used to compute groupwise metrics

    Returns
    -------
    `pd.Series` object
        Series object containing the ranking metrics
        computed using the list of secondary labels
        on the old and new ranks generated by the model
    Returns
    -------
    """
    secondary_labels_metrics_dict = {
        k: v[0] for k, v in query_group[group_keys].to_dict(orient="list").items()
    }
    for secondary_label in secondary_labels:
        try:
            # Compute failure stats for before and after ranking with model
            secondary_labels_metrics_dict.update(
                compute_secondary_label_metrics(
                    secondary_label_values=query_group[secondary_label],
                    ranks=query_group[old_rank_col],
                    click_rank=query_group[query_group[label_col] == 1][old_rank_col].values[0],
                    secondary_label=secondary_label,
                    prefix="old_",
                )
            )
            secondary_labels_metrics_dict.update(
                compute_secondary_label_metrics(
                    secondary_label_values=query_group[secondary_label],
                    ranks=query_group[new_rank_col],
                    click_rank=query_group[query_group[label_col] == 1][new_rank_col].values[0],
                    secondary_label=secondary_label,
                    prefix="new_",
                )
            )
        except IndexError:
            # Ignore queries with no/invalid click ranks
            continue

    return pd.Series(secondary_labels_metrics_dict)


def get_grouped_stats(
    df: pd.DataFrame,
    query_key_col: str,
    label_col: str,
    old_rank_col: str,
    new_rank_col: str,
    group_keys: List[str] = [],
    secondary_labels: List[str] = [],
):
    """
    Compute query stats that can be used to compute ranking metrics

    Parameters
    ----------
    df : `pd.DataFrame` object
        DataFrame object to compute stats and metrics on
    query_key_col : str
        Name of the query key column
    label_col : str
        Name of the label column
    old_rank_col : str
        Name of the column that represents the original rank of the records
    new_rank_col : str
        Name of the column that represents the newly computed rank of the records
        after reordering based on new model scores
    group_keys : list, optional
        List of features used to compute groupwise metrics
    secondary_labels : list, optional
        List of features used to compute failure metrics

    Returns
    -------
    `pd.DataFrame` object
        DataFrame object containing the ranking stats and failure stats
        computed from the old and new ranks and secondary labels generated
        by the model
    """

    # Compute metrics on secondary labels
    df_secondary_labels_metrics = pd.DataFrame()
    if secondary_labels:
        df_secondary_labels_metrics = df.groupby(query_key_col).apply(
            lambda grp: compute_secondary_labels_metrics_on_query_group(
                query_group=grp,
                label_col=label_col,
                old_rank_col=old_rank_col,
                new_rank_col=new_rank_col,
                secondary_labels=secondary_labels,
                group_keys=group_keys
            ))

    # Select clicked records
    df_clicked = df[df[label_col] == 1.0]

    if group_keys:
        df_grouped_batch = df_clicked.groupby(group_keys)

        # Compute groupwise stats
        query_count = df_grouped_batch[query_key_col].nunique()
        sum_old_rank = df_grouped_batch.apply(lambda x: x[old_rank_col].sum())
        sum_new_rank = df_grouped_batch.apply(lambda x: x[new_rank_col].sum())
        sum_old_reciprocal_rank = df_grouped_batch.apply(lambda x: (1.0 / x[old_rank_col]).sum())
        sum_new_reciprocal_rank = df_grouped_batch.apply(lambda x: (1.0 / x[new_rank_col]).sum())

        # Aggregate secondary label metrics by group keys
        df_secondary_labels_metrics = df_secondary_labels_metrics.groupby(group_keys).sum()
    else:
        # Compute overall stats if group keys are not specified
        query_count = [df_clicked.shape[0]]
        sum_old_rank = [df_clicked[old_rank_col].sum()]
        sum_new_rank = [df_clicked[new_rank_col].sum()]
        sum_old_reciprocal_rank = [(1.0 / df_clicked[old_rank_col]).sum()]
        sum_new_reciprocal_rank = [(1.0 / df_clicked[new_rank_col]).sum()]

        # Aggregate secondary label metrics
        df_secondary_labels_metrics = df_secondary_labels_metrics.sum().to_frame().T

    df_label_stats = pd.DataFrame(
        {
            "query_count": query_count,
            "sum_old_rank": sum_old_rank,
            "sum_new_rank": sum_new_rank,
            "sum_old_reciprocal_rank": sum_old_reciprocal_rank,
            "sum_new_reciprocal_rank": sum_new_reciprocal_rank,
        }
    )

    df_stats = pd.concat([df_label_stats, df_secondary_labels_metrics], axis=1)

    return df_stats


def summarize_grouped_stats(df_grouped):
    """
    Summarize and compute metrics from grouped ranking data stats

    Parameters
    ----------
    df_grouped : `pd.DataFrame` object
        Grouped DataFrame object containing the query level stats
        to use for computing ranking metrics

    Returns
    -------
    `pd.DataFrame` object
        DataFrame object containing ranking metrics computed from input
        query stats dataframe
    """

    if isinstance(df_grouped, pd.Series):
        df_grouped_metrics = df_grouped.to_frame().T.sum()
    else:
        df_grouped_metrics = df_grouped.sum()

    query_count = df_grouped_metrics["query_count"]
    df_grouped_metrics = df_grouped_metrics / query_count
    df_grouped_metrics["query_count"] = query_count

    # Rename metrics appropriately
    df_grouped_metrics = df_grouped_metrics.rename(
        {c: "{}_mean".format(c) for c in df_grouped_metrics.to_dict().keys()}
    )
    df_grouped_metrics = df_grouped_metrics.rename(
        {
            "sum_old_rank_mean": "old_ACR",
            "sum_new_rank_mean": "new_ACR",
            "sum_old_reciprocal_rank_mean": "old_MRR",
            "sum_new_reciprocal_rank_mean": "new_MRR",
            "query_count_mean": "query_count",
        }
    )

    processed_metric_name_suffixes = set()
    for metric_name in df_grouped_metrics.to_dict().keys():
        metric_name_suffix = metric_name[4:]
        perc_improv_metric_name = "perc_improv_{}".format(metric_name_suffix)

        if metric_name_suffix in processed_metric_name_suffixes:
            continue

        # If higher values of the metric are better/desirable
        if metric_name_suffix.endswith(tuple(POSITIVE_METRIC_SUFFIXES)):
            df_grouped_metrics[perc_improv_metric_name] = 100. * (
                (df_grouped_metrics["new_{}".format(metric_name_suffix)] -
                 df_grouped_metrics["old_{}".format(metric_name_suffix)])
                / df_grouped_metrics["old_{}".format(metric_name_suffix)])

        # If lower values of the metric are better/desirable
        elif metric_name_suffix.endswith(tuple(NEGATIVE_METRIC_SUFFIXES)):
            df_grouped_metrics[perc_improv_metric_name] = 100. * (
                (df_grouped_metrics["old_{}".format(metric_name_suffix)] -
                 df_grouped_metrics["new_{}".format(metric_name_suffix)])
                / df_grouped_metrics["old_{}".format(metric_name_suffix)])

        processed_metric_name_suffixes.add(metric_name_suffix)

    return df_grouped_metrics



import unittest
import tensorflow as tf
import logging

from ml4ir.base.data.tfrecord_reader import TFRecordExampleParser
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.config.keys import TFRecordTypeKey
from ml4ir.base.io.local_io import LocalIO
from ml4ir.base.features.preprocessing import PreprocessingMap

DATASET_PATH = "ml4ir/applications/classification/tests/data/tfrecord/train/file_0.tfrecord"
FEATURE_CONFIG_PATH = "ml4ir/applications/classification/tests/data/configs/feature_config.yaml"


class TFRecordExampleParserTest(unittest.TestCase):
    """
    Test class for ml4ir.base.data.tfrecord_reader.TFRecordExampleParser
    """

    def setUp(self):
        file_io = LocalIO()
        logger = logging.getLogger()

        self.dataset = tf.data.TFRecordDataset(DATASET_PATH)
        self.proto = next(iter(self.dataset))
        self.feature_config = FeatureConfig.get_instance(
            tfrecord_type=TFRecordTypeKey.EXAMPLE,
            feature_config_dict=file_io.read_yaml(FEATURE_CONFIG_PATH),
            logger=logger,
        )
        self.parser = TFRecordExampleParser(
            feature_config=self.feature_config,
            preprocessing_map=PreprocessingMap(),
            required_fields_only=False,
        )

    def test_features_spec(self):
        """
        Test the feature specification constructed and used to parse the Example proto
        """
        features_spec = self.parser.features_spec

        assert isinstance(features_spec, dict)

        # Check if the feature specification matches with the feature_config
        assert len(set(self.feature_config.get_all_features("name"))) == len(features_spec)
        for feature in self.feature_config.get_all_features("name"):
            assert feature in features_spec

    def test_extract_features_from_proto(self):
        """
        Test extraction of features from serialized proto
        """
        features = self.parser.extract_features_from_proto(self.proto)

        for feature, feature_tensor in features.items():
            # Test that no feature is a sparse tensor
            assert not isinstance(feature_tensor, tf.sparse.SparseTensor)

            # Test that each extracted feature is a scalar
            assert feature_tensor.shape == ()

        # Assert that there is no mask feature
        assert "mask" not in features

    def test_get_default_tensor(self):
        """
        Test the default tensor used for missing features
        """
        default_tensor = self.parser.get_default_tensor(
            self.feature_config.get_feature("query_text")
        )
        assert default_tensor.shape == ()

        default_tensor = self.parser.get_default_tensor(
            self.feature_config.get_feature("user_context")
        )
        assert default_tensor.shape == ()

    def test_get_feature(self):
        """
        Test fetching feature tensor from extracted feature dictionary
        """
        feature_tensor = self.parser.get_feature(
            self.feature_config.get_feature("query_text"), {"query_text": tf.zeros((3, 4, 6))}
        )
        assert feature_tensor.shape == (1, 3, 4, 6)

        # Check missing feature being replaced with default tensor
        feature_tensor = self.parser.get_feature(self.feature_config.get_feature("query_text"), {})
        assert feature_tensor.shape == (1,)

    def test_generate_and_add_mask(self):
        """
        Test mask generation and addition
        """
        features_dict, sequence_size = self.parser.generate_and_add_mask({}, {})

        assert "mask" not in features_dict
        assert sequence_size == tf.constant(0)

    def test_pad_feature(self): 
        """ 
        Test feature padding to max sequence size   
        """ 
        feature_tensor = self.parser.pad_feature(tf.zeros((10)),    
                                                 self.feature_config.get_feature("query_text")) 

        # Check that there was no padding done  
        assert feature_tensor.shape == (10,)

    def test_parse_fn(self):
        """
        Test the Example parsing function
        """
        features, labels = self.parser.get_parse_fn()(self.proto)

        assert isinstance(features, dict)
        assert isinstance(labels, tf.Tensor)

        for feature in self.feature_config.get_all_features(key="node_name", include_label=False):
            assert feature in features

        # Check tensor shapes
        assert features["query_key"].shape == (1,)
        assert features["query_text"].shape == (1,)
        assert features["query_words"].shape == (1, 20)
        assert features["domain_id"].shape == (1,)
        assert features["user_context"].shape == (1, 20)
        assert labels.shape == (1,)



from ml4ir.base.config.keys import Key


class LossKey(Key):
    """Model loss keys that can be used with a ranking model"""

    CATEGORICAL_CROSS_ENTROPY = "categorical_cross_entropy"
    SIGMOID_CROSS_ENTROPY = "sigmoid_cross_entropy"
    RANK_ONE_LISTNET = "rank_one_listnet"
    SOFTMAX_CROSS_ENTROPY = "softmax_cross_entropy"


class ScoringTypeKey(Key):
    """Scoring keys"""

    POINTWISE = "pointwise"
    PAIRWISE = "pairwise"
    GROUPWISE = "groupwise"
    LISTWISE = "listwise"


class MetricKey(Key):
    """Model metric keys that can be used with a ranking model"""

    MRR = "MRR"
    ACR = "ACR"
    NDCG = "NDCG"
    PRECISION = "Precision"
    CATEGORICAL_ACCURACY = "categorical_accuracy"
    TOP_5_CATEGORICAL_ACCURACY = "top_5_categorical_accuracy"


class LossTypeKey(Key):
    """Loss type keys"""

    POINTWISE = "pointwise"
    PAIRWISE = "pairwise"
    LISTWISE = "listwise"


class PositionalBiasHandler(Key):
    """Types of models to handle positional bias"""

    FIXED_ADDITIVE_POSITIONAL_BIAS = "fixed_additive_positional_bias"



import string

from ml4ir.applications.ranking.tests.test_base import RankingTestBase
from ml4ir.base.features import preprocessing

import tensorflow as tf
import numpy as np

class RankingModelTest(RankingTestBase):
    def test_text_preprocesing(self):
        """
        Asserts the preprocessing of a string tensor by
        converting it to its lower case form and removing punctuations
        """
        input_text = "ABCabc123!@#"
        processed_text = (
            preprocessing.preprocess_text(input_text, remove_punctuation=True, to_lower=True)
            .numpy()
            .decode("utf-8")
        )

        # Converting to lower case
        assert processed_text.lower() == processed_text

        # Removing punctuation
        assert (
            processed_text.translate(str.maketrans("", "", string.punctuation)) == processed_text
        )

        # Overall
        assert processed_text.replace("\x00", "") == input_text.lower().translate(
            str.maketrans("", "", string.punctuation)
        )
        assert processed_text.replace("\x00", "") == "abcabc123"

    def test_text_preprocesing_with_replace_by_whitespace(self):
        """
        Asserts the preprocessing of a string tensor with custom punctuation character and whitespace replacement character
        """
        input_text = " # abc. bcd-$#efg@hij ."
        processed_text = (
            preprocessing.preprocess_text(input_text,
                                          remove_punctuation=True,
                                          to_lower=True,
                                          punctuation=".-$#",
                                          replace_with_whitespace=True)
                .numpy()
                .decode("utf-8")
        )

        self.assertEqual("abc bcd efg@hij", processed_text)

    def test_get_one_hot_vectorizer(self):
        """
        Asserts ml4ir.base.features.preprocessing.get_one_hot_vectorizer
        """
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "fn": "categorical_indicator_with_vocabulary_file",
                "args": {
                    "vocabulary_file": "ml4ir/applications/classification/tests/data/configs/vocabulary/entity_id.csv",
                    "num_oov_buckets": 1,
                },
            },
            "default_value": "",
        }

        one_hot_vectorizer = preprocessing.get_one_hot_label_vectorizer(feature_info, self.file_io)

        # Assert 1st position
        one_hot_labels = one_hot_vectorizer(tf.constant(["AAA"]))
        expected_one_hot_labels = tf.constant([[1.] + 8*[0.]])
        assert tf.reduce_all(tf.equal(one_hot_labels, expected_one_hot_labels))

        # Assert 7th position
        one_hot_labels = one_hot_vectorizer(tf.constant(["GGG"]))
        expected_one_hot_labels = tf.constant([6*[0.] + [1.] + 2*[0.]])
        assert tf.reduce_all(tf.equal(one_hot_labels, expected_one_hot_labels))

        # Assert last position for out of vocabulary
        one_hot_labels = one_hot_vectorizer(tf.constant(["out of vocabulary"]))
        expected_one_hot_labels = tf.constant([8*[0.] + [1.]])
        assert tf.reduce_all(tf.equal(one_hot_labels, expected_one_hot_labels))

    def test_split_and_pad_string(self):
        """
        Asserts ml4ir.base.features.preprocessing.split_and_pad_string
        """
        input_text = tf.constant(["_A_BC_ab_c1_23!@_#"])

        split_text = preprocessing.split_and_pad_string(input_text, split_char="_", max_length=25)
        expected_split_text = tf.constant(["", "A", "BC", "ab", "c1", "23!@", "#"] + [""]*18)

        assert tf.reduce_all(tf.equal(split_text, expected_split_text))

    def testing_click_conversion(self):
        typ = 'int'
        label_vector = np.ones(10, dtype=typ)
        label_vector = tf.convert_to_tensor(label_vector)
        clicks = preprocessing.convert_label_to_clicks(label_vector, typ)
        comp = tf.equal(label_vector, clicks)
        assert sum(tf.dtypes.cast(comp, 'int8')) == 10

        typ = 'float'
        label_vector = np.ones(10, dtype=typ)
        label_vector[0] = 5
        label_vector[-1] = 5
        label_vector = tf.convert_to_tensor(label_vector)
        clicks = preprocessing.convert_label_to_clicks(label_vector, typ)
        assert clicks[0] == 1 and clicks[-1] == 1 and sum(clicks[1:-1]) == 0

        typ = 'int'
        label_vector = np.zeros(10, dtype=typ)
        label_vector = tf.convert_to_tensor(label_vector)
        clicks = preprocessing.convert_label_to_clicks(label_vector, typ)
        assert sum(tf.dtypes.cast(comp, 'int8')) == 10



import unittest
import tensorflow as tf
import numpy as np
import os
import random
import gc
from argparse import Namespace
from tensorflow.keras.metrics import Metric
from tensorflow.keras.optimizers import Optimizer

from ml4ir.base.model.relevance_model import RelevanceModel
from ml4ir.base.model.losses.loss_base import RelevanceLossBase
from ml4ir.base.model.scoring.scoring_model import ScorerBase, RelevanceScorer
from ml4ir.base.model.scoring.interaction_model import InteractionModel, UnivariateInteractionModel
from ml4ir.base.model.optimizers.optimizer import get_optimizer
from ml4ir.base.io.local_io import LocalIO
from ml4ir.base.io.logging_utils import setup_logging
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.config.keys import ArchitectureKey
from ml4ir.applications.ranking.model.ranking_model import RankingModel, LinearRankingModel
from ml4ir.applications.ranking.model.losses import loss_factory
from ml4ir.applications.ranking.model.metrics import metric_factory
from ml4ir.applications.ranking.config.parse_args import get_args

import warnings
from typing import Union, List, Type

warnings.filterwarnings("ignore")


OUTPUT_DIR = "ml4ir/applications/ranking/tests/test_output"
ROOT_DATA_DIR = "ml4ir/applications/ranking/tests/data"
FEATURE_CONFIG_FNAME = "feature_config.yaml"


class RelevanceTestBase(unittest.TestCase):
    """
    This is the base test class for the common relevance code under ml4ir/base/

    Inherit this class to define tests which need the default pipeline args and configs.
    """
    def setUp(
        self,
        output_dir: str = OUTPUT_DIR,
        root_data_dir: str = ROOT_DATA_DIR,
        feature_config_fname: str = FEATURE_CONFIG_FNAME,
    ):
        self.output_dir = output_dir
        self.root_data_dir = root_data_dir
        self.feature_config_fname = feature_config_fname
        self.file_io = LocalIO()

        # Make temp output directory
        self.file_io.make_directory(self.output_dir, clear_dir=True)

        # Fix random seed values for repeatability
        tf.keras.backend.clear_session()
        np.random.seed(123)
        tf.random.set_seed(123)
        random.seed(123)

        # Setup arguments
        self.args: Namespace = get_args([])
        self.args.models_dir = output_dir
        self.args.logs_dir = output_dir

        self.load_model_config(self.args.model_config)

        # Setup logging
        outfile: str = os.path.join(self.args.logs_dir, "output_log.csv")

        self.logger = setup_logging(reset=True, file_name=outfile, log_to_file=True)

    def tearDown(self):
        # Delete output directory
        self.file_io.rm_dir(self.output_dir)

        # Delete other temp directories
        self.file_io.rm_dir(os.path.join(self.root_data_dir, "csv", "tfrecord"))

        # Clear memory
        tf.keras.backend.clear_session()
        gc.collect()

    def load_model_config(self, model_config_path:str):
        """Load the model config dictionary"""
        self.model_config = self.file_io.read_yaml(model_config_path)

if __name__ == "__main__":
    unittest.main()



import os
import numpy as np
import random
import tensorflow as tf
import glob

from ml4ir.base.config.keys import DataFormatKey
from ml4ir.applications.ranking.config.keys import MetricKey
from ml4ir.applications.ranking.tests.test_base import RankingTestBase
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.applications.ranking.model.ranking_model import RankingModel
from ml4ir.base.features.feature_config import FeatureConfig


class RankingTransferLearningTest(RankingTestBase):
    """
    This class defines the unit tests around the transfer learning
    capability of RelevanceModel.

    Transfer learning in ml4ir is enabled by saving individual layer weights
    at training time. When defining a new RelevanceModel, these pretrained
    weights can be loaded by passing a mapping of layer name to the path of
    the pretrained weights(saved previously). Additionally, these weights can
    be frozen to avoid fine tuning as needed.
    """

    def get_ranking_dataset_and_model(
        self, seed=123, initialize_layers_dict={}, freeze_layers_list=[]
    ):
        """Helper method to get a RankingModel and Dataset with some default args"""
        data_dir = os.path.join(self.root_data_dir, DataFormatKey.TFRECORD)
        feature_config_path = os.path.join(self.root_data_dir, "configs", self.feature_config_fname)
        data_format = DataFormatKey.TFRECORD
        metrics_keys = [MetricKey.MRR]

        # Fix random seed values for repeatability
        tf.keras.backend.clear_session()
        np.random.seed(seed)
        tf.random.set_seed(seed)
        random.seed(seed)

        feature_config: FeatureConfig = FeatureConfig.get_instance(
            tfrecord_type=self.args.tfrecord_type,
            feature_config_dict=self.file_io.read_yaml(feature_config_path),
            logger=self.logger,
        )

        relevance_dataset = RelevanceDataset(
            data_dir=data_dir,
            data_format=data_format,
            feature_config=feature_config,
            tfrecord_type=self.args.tfrecord_type,
            max_sequence_size=self.args.max_sequence_size,
            batch_size=self.args.batch_size,
            preprocessing_keys_to_fns={},
            train_pcent_split=self.args.train_pcent_split,
            val_pcent_split=self.args.val_pcent_split,
            test_pcent_split=self.args.test_pcent_split,
            use_part_files=self.args.use_part_files,
            parse_tfrecord=True,
            file_io=self.file_io,
            logger=self.logger,
        )

        ranking_model: RankingModel = self.get_ranking_model(
            loss_key=self.args.loss_key,
            feature_config=feature_config,
            metrics_keys=metrics_keys,
            initialize_layers_dict=initialize_layers_dict,
            freeze_layers_list=freeze_layers_list,
        )

        return ranking_model, relevance_dataset

    def test_model_saving(self):
        """
        This unit test checks if the individual layer weights were saved as part of the saved model
        """
        model, dataset = self.get_ranking_dataset_and_model(seed=123)
        model.save(
            models_dir=self.args.models_dir,
            preprocessing_keys_to_fns={},
            postprocessing_fn=None,
            required_fields_only=not self.args.use_all_fields_at_inference,
            pad_sequence=self.args.pad_sequence_at_inference,
        )

        # Check that both default and tfrecord serving signatures have been saved
        saved_directories = [
            os.path.basename(d)
            for d in glob.glob(os.path.join(self.args.models_dir, "final", "*"))
        ]
        assert "default" in saved_directories
        assert "tfrecord" in saved_directories

        # Check if all model layers have been saved
        assert "layers" in saved_directories
        saved_layers = [
            os.path.basename(f)
            for f in glob.glob(os.path.join(self.args.models_dir, "final", "layers", "*"))
        ]
        for layer in model.model.layers:
            assert "{}.npz".format(layer.name.split("/")[0]) in saved_layers

    def test_loading_pretrained_weights(self):
        """
        This unit test checks if the pretrained weights were loaded into the RankingModel
        """
        pretrained_layer_path = os.path.join(
            self.root_data_dir, "models", "layers", "query_text_bytes_embedding.npz"
        )
        model, dataset = self.get_ranking_dataset_and_model(
            seed=125, initialize_layers_dict={"query_text_bytes_embedding": pretrained_layer_path}
        )

        model_layer_weights = model.model.get_layer("query_text_bytes_embedding").get_weights()
        pretrained_layer_weights = self.file_io.load_numpy_array(pretrained_layer_path)

        for i in range(len(model_layer_weights)):
            assert np.all(model_layer_weights[i] == pretrained_layer_weights[i])

    def test_freeze_weights(self):
        """
        This unit test checks if weights for a layer were frozen at training time
        """
        layer_name = "query_text_bytes_embedding"
        model, dataset = self.get_ranking_dataset_and_model(
            seed=123, freeze_layers_list=[layer_name]
        )

        # Assert non trainable weights
        initial_layer = model.model.get_layer(layer_name)
        assert not initial_layer.trainable
        assert len(initial_layer.trainable_weights) == 0

        # Train the model for 1 epoch and test that the frozen layer weights remain same
        model.fit(dataset=dataset, num_epochs=1, models_dir=self.output_dir)
        final_layer = model.model.get_layer(layer_name)

        for i in range(len(initial_layer.get_weights())):
            assert np.all(initial_layer.get_weights()[i] == final_layer.get_weights()[i])



import numpy as np
from scipy import stats
import unittest
import warnings
from ml4ir.applications.ranking.t_test import perform_click_rank_dist_paired_t_test, compute_stats_from_stream

warnings.filterwarnings("ignore")

np.random.seed(123)


class TestTtestCalculation(unittest.TestCase):
    """
    Testing the calculation of paired t-test statistic and its p-value.
    """

    def t_test_calculation(self, a_bucket, b_bucket, batch_size):
        n = len(a_bucket)
        expected_t_test_stat, expected_pvalue = stats.ttest_rel(a_bucket, b_bucket)
        agg_mean, agg_count, agg_M2 = 0, 0, 0
        for bcount in range(int(n/batch_size)):
            batch_a = a_bucket[bcount*batch_size : (bcount+1)*(batch_size)]
            batch_b = b_bucket[bcount*batch_size : (bcount+1)*(batch_size)]
            d = batch_a - batch_b
            agg_count, agg_mean, agg_M2 = compute_stats_from_stream(d, agg_count, agg_mean, agg_M2)

        t_test_stat, pvalue = perform_click_rank_dist_paired_t_test(agg_mean, (agg_M2/(agg_count-1)), agg_count)
        assert np.isclose(expected_pvalue, pvalue, atol=0.0001)

    def test_1(self):
        n = 100
        batch_size = 10
        batch_a = np.random.randn(n) + 1
        batch_b = np.random.randn(n)
        self.t_test_calculation(batch_a, batch_b, batch_size)

    def test_2(self):
        n = 10
        batch_size = 3
        batch_a = np.random.randn(n) + 20
        batch_b = np.random.randn(n)
        self.t_test_calculation(batch_a, batch_b, batch_size)

    def test_3(self):
        n = 50
        batch_size = 5
        batch_a = np.random.randn(n)
        batch_b = np.array(batch_a)
        batch_b[0] += 1
        self.t_test_calculation(batch_a, batch_b, batch_size)

    def test_4(self):
        n = 50
        batch_size = 5
        batch_a = np.random.randn(n)
        batch_b = np.array(batch_a)
        batch_b[0] += 1
        self.t_test_calculation(batch_a, batch_b, batch_size)

    def test_5(self):
        n = 100
        batch_size = 5
        batch_a = np.random.randn(n)
        batch_b = np.random.randn(n)
        self.t_test_calculation(batch_a, batch_b, batch_size)

    def test_6(self):
        n = 100
        batch_size = 5
        batch_a = np.random.exponential(scale=0.1, size=n)
        batch_b = np.random.exponential(scale=0.5, size=n)
        self.t_test_calculation(batch_a, batch_b, batch_size)
        
        
    def test_6(self):
        n = 100
        batch_size = 5
        batch_a = np.random.exponential(scale=0.1, size=n)
        batch_b = np.random.exponential(scale=0.5, size=n)
        self.t_test_calculation(batch_a, batch_b, batch_size)


    def test_7(self):
        n = 10
        batch_size = 2
        batch_a = np.random.exponential(scale=1.0, size=n)
        batch_b = np.random.exponential(scale=0.2, size=n)
        self.t_test_calculation(batch_a, batch_b, batch_size)


    def test_7(self):
        n = 10
        batch_size = 2
        batch_a = np.random.exponential(scale=1.0, size=n)
        batch_b = np.random.exponential(scale=0.2, size=n)
        self.t_test_calculation(batch_a, batch_b, batch_size)

    def test_8(self):
        n = 50
        batch_size = 10
        batch_a = np.random.exponential(scale=0.1, size=n)
        batch_b = np.random.exponential(scale=0.1, size=n)
        self.t_test_calculation(batch_a, batch_b, batch_size)

    def test_8(self):
        n = 50
        batch_size = 10
        batch_a = np.random.exponential(scale=0.1, size=n)
        batch_b = np.random.exponential(scale=0.1, size=n)
        self.t_test_calculation(batch_a, batch_b, batch_size)

    def test_9(self):
        n = 50
        batch_size = 5
        batch_a = np.random.exponential(scale=0.1, size=n)
        batch_b = batch_a + 0.001
        self.t_test_calculation(batch_a, batch_b, batch_size)


if __name__ == "__main__":
    unittest.main()



from scipy import stats
import numpy as np


def perform_click_rank_dist_paired_t_test(mean, variance, n):
    """
    Compute the paired t-test statistic and its p-value given mean, standard deviation and sample count
    Parameters
    ----------
    mean: float
        The mean of the rank differences for the entire dataset
    variance: float
        The variance of the rank differences for the entire dataset
    n: int
        The number of samples in the entire dataset

    Returns
    -------
    t_test_stat: float
        The t-test statistic
    pvalue: float
        The p-value of the t-test statistic
    """
    t_test_stat = mean / (np.sqrt(variance/n))
    df = n - 1
    pvalue = 1 - stats.t.cdf(np.abs(t_test_stat), df=df)
    return t_test_stat*2, pvalue*2  # multiplying by 2 for two sided t-test


def compute_stats_from_stream(diff, count, mean, m2):
    """
    Compute the running mean, variance for a stream of data.
    src: Welford's online algorithm: https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance

    Parameters
    ----------
    diff: List
        A batch of differences in rankings
    count: int
        Aggregates the number of samples seen so far
    mean: float
        Accumulates the mean of the rank differences so far
    m2: float
        Aggregates the squared distance from the mean

    Returns
    -------
    count: int
        The updated aggregate sample count
    mean: float
        The updated mean of the rank differences so far
    m2: float
        The updated squared distance from the mean
    """
    for i in range(len(diff)):
        count += 1
        delta = diff[i] - mean
        mean += delta / count
        delta2 = diff[i] - mean
        m2 += delta * delta2
    return count, mean, m2


def t_test_log_results(t_test_stat, pvalue, ttest_pvalue_threshold, logger):
    """
    performing click rank distribution t-test

    Parameters
    ----------
    t_test_stat: float
        The t-test statistic
    pvalue: float
        The p-value of the t-test statistic
    ttest_pvalue_threshold: float
        The p-value threshold
    logger: Logger
        Logger object to log t-test significance decision

    """
    logger.info(
        "Performing a paired t-test between the click rank distribution of new model and the old model:\n\tNull hypothesis: There is no difference between the two click distributions.\n\tAlternative hypothesis: There is a difference between the two click distributions")
    logger.info("t-test statistic={}, p-value={}".format(t_test_stat, pvalue))
    if pvalue < ttest_pvalue_threshold:
       logger.info(
            "With p-value threshold={} > p-value --> we reject the null hypothesis. The click rank distribution of the new model is significantly different from the old model".format(
                ttest_pvalue_threshold))
    else:
       logger.warning(
            "With p-value threshold={} < p-value --> we cannot reject the null hypothesis. The click rank distribution of the new model is not significantly different from the old model".format(
                ttest_pvalue_threshold))



"""
Writes data in Example or SequenceExample protobuf (tfrecords) format.

To use it as a standalone script, refer to the argument spec
at the bottom

Notes
-----

Setting ``--keep-single-files`` writes one tfrecord file
for each CSV file (better performance). If not set,
joins everything to a single tfrecord file.

Examples
--------

Syntax to convert a single or several CSVs:

>>> python ml4ir/base/data/tfrecord_writer.py \\
... sequence_example|example \\
... --csv-files <SPACE_SEPARATED_PATHS_TO_CSV_FILES> \\
... --out-dir <PATH_TO_OUTPUT_DIR> \\
... --feature_config <PATH_TO_YAML_FEATURE_CONFIG> \\
... --keep-single-files

or to convert all CSV files in a dir

>>> python ml4ir/base/data/tfrecord_writer.py \\
... sequence_example|example \\
... --csv-dir <DIR_WITH_CSVs> \\
... --out-dir <PATH_TO_OUTPUT_DIR> \\
... --feature_config <PATH_TO_YAML_FEATURE_CONFIG> \\
... --keep-single-files

Usage example:

>>> python ml4ir/base/data/tfrecord_writer.py \\
... sequence_example \\
... --csv-files /tmp/d.csv /tmp/d2.csv \\
... --out-dir /tmp \\
... --feature-config /tmp/fconfig.yaml \\
... --keep-single-files
 """

from tensorflow import io
from typing import List
from logging import Logger
from argparse import ArgumentParser
import os
from pandas import DataFrame

from ml4ir.base.io.file_io import FileIO
from ml4ir.base.io.local_io import LocalIO
from ml4ir.base.config.keys import TFRecordTypeKey
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.io.logging_utils import setup_logging
from ml4ir.base.data.tfrecord_helper import get_sequence_example_proto, get_example_proto

MODES = {"example": TFRecordTypeKey.EXAMPLE, "sequence_example": TFRecordTypeKey.SEQUENCE_EXAMPLE}


def write_from_files(
    csv_files: List[str],
    tfrecord_file: str,
    feature_config: FeatureConfig,
    tfrecord_type: str,
    file_io: FileIO,
    logger: Logger = None,
):
    """
    Converts data from CSV files into tfrecord files

    Parameters
    ----------
    csv_files : list of str
        list of csv file paths to read data from
    tfrecord_file : str
        tfrecord file path to write the output
    feature_config : `FeatureConfig`
        FeatureConfig object that defines the features to be loaded in the dataset
        and the preprocessing functions to be applied to each of them
    tfrecord_type : {"example", "sequence_example"}
        Type of the TFRecord protobuf message to be used for TFRecordDataset
    logger : `Logger`, optional
        logging handler for status messages
    """

    # Read CSV data into a pandas dataframe
    df = file_io.read_df_list(csv_files)
    write_from_df(df, tfrecord_file, feature_config, tfrecord_type, logger)


def write_from_df(
    df: DataFrame,
    tfrecord_file: str,
    feature_config: FeatureConfig,
    tfrecord_type: str,
    logger: Logger = None,
):
    """
    Converts data from CSV files into tfrecord files

    Parameters
    df : `pd.DataFrame`
        pandas DataFrame to be converted to TFRecordDataset
    tfrecord_file : str
        tfrecord file path to write the output
    feature_config : `FeatureConfig`
        FeatureConfig object that defines the features to be loaded in the dataset
        and the preprocessing functions to be applied to each of them
    tfrecord_type : {"example", "sequence_example"}
        Type of the TFRecord protobuf message to be used for TFRecordDataset
    logger : `Logger`, optional
        logging handler for status messages
    """

    if logger:
        logger.info("Writing SequenceExample protobufs to : {}".format(tfrecord_file))
    with io.TFRecordWriter(tfrecord_file) as tf_writer:
        if tfrecord_type == TFRecordTypeKey.EXAMPLE:
            protos = df.apply(
                lambda row: get_example_proto(row=row, features=feature_config.get_all_features()),
                axis=1,
            )
        elif tfrecord_type == TFRecordTypeKey.SEQUENCE_EXAMPLE:
            # Group pandas dataframe on query_id/query key and
            # convert each group to a single sequence example proto
            context_feature_names = feature_config.get_context_features(key="name")
            protos = df.groupby(context_feature_names).apply(
                lambda g: get_sequence_example_proto(
                    group=g,
                    context_features=feature_config.get_context_features(),
                    sequence_features=feature_config.get_sequence_features(),
                )
            )
        else:
            raise Exception(
                "You have entered {} as tfrecords write mode. "
                "We only support {} and {}.".format(
                    tfrecord_type, TFRecordTypeKey.EXAMPLE, TFRecordTypeKey.SEQUENCE_EXAMPLE
                )
            )
        # Write to disk
        for proto in protos:
            tf_writer.write(proto.SerializeToString())


def main(args):
    """Convert CSV files into tfrecord Example/SequenceExample files"""
    # Setup logging
    logger: Logger = setup_logging()
    file_io = LocalIO(logger)

    # Get all CSV files to be converted, depending on user's arguments
    if args.csv_dir:
        csv_files: List[str] = file_io.get_files_in_directory(
            indir=args.csv_dir, extension="*.csv"
        )
    else:
        csv_files: List[str] = args.csv_files

    # Load feat config

    feature_config: FeatureConfig = FeatureConfig.get_instance(
        tfrecord_type=MODES[args.tfmode],
        feature_config_dict=file_io.read_yaml(args.feature_config),
        logger=logger,
    )

    # Convert to TFRecord SequenceExample protobufs and save
    if args.keep_single_files:
        # Convert each CSV file individually - better performance
        for csv_file in csv_files:
            tfrecord_file: str = os.path.basename(csv_file).replace(".csv", "")
            tfrecord_file: str = os.path.join(args.out_dir, "{}.tfrecord".format(tfrecord_file))
            write_from_files(
                csv_files=[csv_file],
                tfrecord_file=tfrecord_file,
                feature_config=feature_config,
                logger=logger,
                tfrecord_type=MODES[args.tfmode],
            )

    else:
        # Convert all CSV files at once - expensive groupby operation
        tfrecord_file: str = os.path.join(args.out_dir, "combined.tfrecord")
        write_from_files(
            csv_files=csv_files,
            tfrecord_file=tfrecord_file,
            feature_config=feature_config,
            logger=logger,
            tfrecord_type=MODES[args.tfmode],
            file_io=file_io,
        )


def define_arguments():
    first_doc_line = __doc__.strip().split("\n")[0]
    parser = ArgumentParser(description=first_doc_line)
    parser.add_argument(
        "tfmode",
        choices=MODES,  # Choices with a dict, shows the keys
        help="select between `sequence` and `example` to write tf.Example or tf.SequenceExample",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--csv-dir", type=str, help="Directory with CSV files; every .csv file will be converted."
    )
    group.add_argument(
        "--csv-files",
        type=str,
        nargs="+",
        help="A single or more (space separated) CSV files to be converted.",
    )
    parser.add_argument(
        "--out-dir",
        type=str,
        default=".",
        help="Output directory for TFRecord files. Base filenames are maintained and .tfrecord is added as extension.",
    )
    parser.add_argument(
        "--feature-config",
        type=str,
        default=None,
        help="Path to feature config YAML file or feature config YAML string",
    )
    parser.add_argument(
        "--keep-single-files",
        action="store_true",
        help="When passed, converts CSV files individually. "
        "Results are written to out-dir replacing the filename's extension with .tfrecord."
        "If not set, a single combined.tfrecord is created."
        "All occurrences of a query key should be within a single file",
    )
    return parser


if __name__ == "__main__":
    my_parser = define_arguments()
    args = my_parser.parse_args()
    main(args)



import socket
import tensorflow as tf
import numpy as np
import pandas as pd
import json
import random
import traceback
import os
import sys
import time
from argparse import Namespace
from logging import Logger
import pathlib
from typing import List
import copy

from ml4ir.base.config.parse_args import get_args
from ml4ir.base.config.dynamic_args import override_with_dynamic_args
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.io import logging_utils
from ml4ir.base.io.local_io import LocalIO
from ml4ir.base.io.spark_io import SparkIO
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.data.kfold_relevance_dataset import KfoldRelevanceDataset
from ml4ir.base.model.relevance_model import RelevanceModel
from ml4ir.base.config.keys import DataFormatKey
from ml4ir.base.config.keys import ExecutionModeKey
from ml4ir.base.config.keys import DefaultDirectoryKey
from ml4ir.base.config.keys import FileHandlerKey
from ml4ir.base.config.keys import CalibrationKey


class RelevancePipeline(object):
    """Base class that defines a pipeline to train, evaluate and save
    a RelevanceModel using ml4ir"""

    def __init__(self, args: Namespace):
        """
        Constructor to create a RelevancePipeline object to train, evaluate
        and save a model on ml4ir.
        This method sets up data, logs, models directories, file handlers used.
        The method also loads and sets up the FeatureConfig for the model training
        pipeline

        Parameters
        ----------
        args: argparse Namespace
            arguments to be used with the pipeline.
            Typically, passed from command line arguments
        """
        self.args = args

        # Generate Run ID
        if len(self.args.run_id) > 0:
            self.run_id: str = self.args.run_id
        else:
            self.run_id = "-".join([socket.gethostname(),
                                    time.strftime("%Y%m%d-%H%M%S")])
        self.start_time = time.time()

        # Setup directories
        self.local_io = LocalIO()
        self.models_dir_hdfs = None
        self.logs_dir_hdfs = None
        self.data_dir_hdfs = None
        if self.args.file_handler == FileHandlerKey.SPARK:
            self.models_dir = os.path.join(self.args.models_dir, self.run_id)
            self.logs_dir = os.path.join(self.args.logs_dir, self.run_id)
            self.data_dir = self.args.data_dir

            self.models_dir_local = os.path.join(
                DefaultDirectoryKey.MODELS, self.run_id)
            self.logs_dir_local = os.path.join(
                DefaultDirectoryKey.LOGS, self.run_id)
            self.data_dir_local = os.path.join(
                DefaultDirectoryKey.TEMP_DATA, os.path.basename(self.data_dir)
            )
        else:
            self.models_dir_local = os.path.join(
                self.args.models_dir, self.run_id)
            self.logs_dir_local = os.path.join(self.args.logs_dir, self.run_id)
            self.data_dir_local = self.args.data_dir

        # Setup logging
        self.local_io.make_directory(self.logs_dir_local, clear_dir=True)
        self.logger: Logger = self.setup_logging()
        self.logger.info(
            "Logging initialized. Saving logs to : {}".format(self.logs_dir_local))
        self.logger.info("Run ID: {}".format(self.run_id))
        self.logger.debug("CLI args: \n{}".format(
            json.dumps(vars(self.args), indent=4)))
        self.local_io.set_logger(self.logger)
        self.local_io.make_directory(self.models_dir_local, clear_dir=False)
        self.model_file = self.args.model_file

        # Set the file handlers and respective setup
        if self.args.file_handler == FileHandlerKey.LOCAL:
            self.file_io = self.local_io
        elif self.args.file_handler == FileHandlerKey.SPARK:
            self.file_io = SparkIO(self.logger)

            # Copy data dir from HDFS to local file system
            self.local_io.make_directory(
                dir_path=DefaultDirectoryKey.TEMP_DATA, clear_dir=True)
            self.file_io.copy_from_hdfs(
                self.data_dir, DefaultDirectoryKey.TEMP_DATA)

            # Copy model_file if present from HDFS to local file system
            if self.model_file:
                self.local_io.make_directory(
                    dir_path=DefaultDirectoryKey.TEMP_MODELS, clear_dir=True
                )
                self.file_io.copy_from_hdfs(
                    self.model_file, DefaultDirectoryKey.TEMP_MODELS)
                self.model_file = os.path.join(
                    DefaultDirectoryKey.TEMP_MODELS, os.path.basename(
                        self.model_file)
                )

        # Setup other arguments
        self.loss_key: str = self.args.loss_key
        self.metrics_keys: List[str] = self.args.metrics_keys
        self.data_format: str = self.args.data_format
        self.tfrecord_type: str = self.args.tfrecord_type

        # RankLib/LibSVM data format specific setup
        if args.data_format == DataFormatKey.RANKLIB:
            try:
                self.non_zero_features_only = self.args.non_zero_features_only
                self.keep_additional_info = self.args.keep_additional_info
            except KeyError:
                self.non_zero_features_only = 0
                self.keep_additional_info = 0
        else:
            self.non_zero_features_only = 0
            self.keep_additional_info = 0

        self.model_file = args.model_file

        # Set random seeds
        self.set_seeds()

        self.logger.info("Running pre-processing step.")
        self.pre_processing_step()
        self.logger.info("Pre-processing step done.")

        # Read/Parse feature_config and model_config YAML
        feature_config_dict = self.file_io.read_yaml(args.feature_config)
        model_config_dict = self.file_io.read_yaml(args.model_config)

        # Customize feature_config and model_config dictionaries
        if "feature_config_custom" in args:
            feature_config_dict = override_with_dynamic_args(
                base_dict=feature_config_dict,
                dynamic_args=args.feature_config_custom)
        if "model_config_custom" in args:
            model_config_dict = override_with_dynamic_args(
                base_dict=model_config_dict,
                dynamic_args=args.model_config_custom)
        self.model_config = model_config_dict

        # Define a FeatureConfig object from loaded YAML
        self.feature_config: FeatureConfig = FeatureConfig.get_instance(
            feature_config_dict=feature_config_dict,
            tfrecord_type=self.tfrecord_type,
            logger=self.logger,
        )

        # Finished initialization
        self.logger.info("Relevance Pipeline successfully initialized!")

    def setup_logging(self) -> Logger:
        """
        Set up the logging utilities for the training pipeline
        Additionally, removes pre existing job status files
        """
        # Remove status file from any previous job at the start of the current job
        for status_file in ["_SUCCESS", "_FAILURE"]:
            self.local_io.rm_file(os.path.join(
                self.logs_dir_local, status_file))

        return logging_utils.setup_logging(
            reset=True,
            file_name=os.path.join(self.logs_dir_local, "output_log.csv"),
            log_to_file=True,
        )

    def set_seeds(self, reset_graph=True):
        """
        Set the random seeds for tensorflow and numpy in order to
        replicate results

        Parameters
        ----------
        reset_graph : bool
            Reset the tensorflow graph and clears the keras session
        """
        if reset_graph:
            tf.keras.backend.clear_session()
            self.logger.info("Tensorflow default graph has been reset")
        np.random.seed(self.args.random_state)
        tf.random.set_seed(self.args.random_state)
        random.seed(self.args.random_state)

    def get_relevance_dataset(self, preprocessing_keys_to_fns={}) -> RelevanceDataset:
        """
        Create RelevanceDataset object by loading train, test data as tensorflow datasets

        Parameters
        ----------
        preprocessing_keys_to_fns : dict of (str, function)
            dictionary of function names mapped to function definitions
            that can now be used for preprocessing while loading the
            TFRecordDataset to create the RelevanceDataset object

        Returns
        -------
        `RelevanceDataset` object
            RelevanceDataset object that can be used for training and evaluating
            the model

        Notes
        -----
        Override this method to create custom dataset objects
        """

        # Prepare Dataset
        relevance_dataset = RelevanceDataset(
            data_dir=self.data_dir_local,
            data_format=self.data_format,
            feature_config=self.feature_config,
            tfrecord_type=self.tfrecord_type,
            max_sequence_size=self.args.max_sequence_size,
            batch_size=self.args.batch_size,
            preprocessing_keys_to_fns=preprocessing_keys_to_fns,
            train_pcent_split=self.args.train_pcent_split,
            val_pcent_split=self.args.val_pcent_split,
            test_pcent_split=self.args.test_pcent_split,
            use_part_files=self.args.use_part_files,
            parse_tfrecord=True,
            file_io=self.local_io,
            logger=self.logger,
            non_zero_features_only=self.non_zero_features_only,
            keep_additional_info=self.keep_additional_info,
        )

        return relevance_dataset

    def get_kfold_relevance_dataset(self, num_folds, include_testset_in_kfold, read_data_sets, preprocessing_keys_to_fns={}) -> RelevanceDataset:
        """
        Create RelevanceDataset object by loading train, test data as tensorflow datasets

        Parameters
        ----------
        num_folds: int
            number of folds in kfold
        include_testset_in_kfold: bool
            whether to include the testset in the folds
        read_data_sets: bool
            whether to call `create_dataset` which reads data from files.
        preprocessing_keys_to_fns : dict of (str, function)
            dictionary of function names mapped to function definitions
            that can now be used for preprocessing while loading the
            TFRecordDataset to create the RelevanceDataset object

        Returns
        -------
        `KfoldRelevanceDataset` object
            RelevanceDataset object that can be used for training and evaluating
            the model

        Notes
        -----
        Override this method to create custom dataset objects
        """

        # Prepare Dataset
        relevance_dataset = KfoldRelevanceDataset(
            data_dir=self.data_dir_local,
            data_format=self.data_format,
            feature_config=self.feature_config,
            tfrecord_type=self.tfrecord_type,
            max_sequence_size=self.args.max_sequence_size,
            batch_size=self.args.batch_size,
            preprocessing_keys_to_fns=preprocessing_keys_to_fns,
            train_pcent_split=self.args.train_pcent_split,
            val_pcent_split=self.args.val_pcent_split,
            test_pcent_split=self.args.test_pcent_split,
            use_part_files=self.args.use_part_files,
            parse_tfrecord=True,
            file_io=self.local_io,
            logger=self.logger,
            non_zero_features_only=self.non_zero_features_only,
            keep_additional_info=self.keep_additional_info,
            num_folds=num_folds,
            include_testset_in_kfold=include_testset_in_kfold,
            read_data_sets=read_data_sets
        )

        return relevance_dataset

    def get_relevance_model(self, feature_layer_keys_to_fns={}) -> RelevanceModel:
        """
        Creates RelevanceModel that can be used for training and evaluating

        Parameters
        ----------
        feature_layer_keys_to_fns : dict of (str, function)
            dictionary of function names mapped to tensorflow compatible
            function definitions that can now be used in the InteractionModel
            as a feature function to transform input features

        Returns
        -------
        `RelevanceModel`
            RelevanceModel that can be used for training and evaluating

        Notes
        -----
        Override this method to create custom loss, scorer, model objects
        """
        raise NotImplementedError

    def create_pipeline_for_kfold(self, args):
        raise NotImplementedError

    def run(self):
        """
        Run the pipeline to train, evaluate and save the model. It also runs the pipeline in kfold cross validation
        mode if specified.

        Returns
        -------
        dict
            Experiment tracking dictionary with metrics and metadata for the run.
            Used for model selection and hyperparameter optimization

        Notes
        -----
        Also populates a experiment tracking dictionary containing
        the metadata, model architecture and metrics generated by the model
        """
        if self.args.kfold <= 1:
            # Run ml4ir without kfold cross validation
            return self.run_pipeline()

        if self.args.include_testset_in_kfold:
            if self.args.kfold < 3:
                raise Exception("Number of folds must be > 2")
        else:
            if self.args.kfold < 2:
                raise Exception("Number of folds must be > 1")

        job_status = "_SUCCESS"
        try:
            args = copy.deepcopy(self.args)

            # reading, parsing the dataset (train, validation, test)
            self.logger.info("Reading datasets ...")
            relevance_dataset = self.get_kfold_relevance_dataset(args.kfold, args.include_testset_in_kfold,
                                                                 read_data_sets=True)
            self.logger.info("Relevance Dataset created")

            merged_data = relevance_dataset.merge_datasets()

            num_folds = self.args.kfold
            base_logs_dir = str(self.args.logs_dir)
            base_models_dir = str(self.args.models_dir)
            base_run_id = self.run_id
            self.logger.info(
                "K-fold Cross Validation mode starting with k={}".format(self.args.kfold))
            self.logger.info("Include testset in the folds={}".format(
                str(self.args.include_testset_in_kfold)))

            # when creating folds, the validation set is assigned fold i, test fold i+1 and training get the rest of folds
            for fold_id in range(num_folds):
                self.logger.info("fold={}".format(fold_id))
                logs_dir = pathlib.Path(base_logs_dir) / self.args.run_id / \
                    "fold_{}".format(fold_id)
                models_dir = pathlib.Path(base_models_dir) / \
                    self.args.run_id / "fold_{}".format(fold_id)
                args.logs_dir = pathlib.Path(logs_dir).as_posix()
                args.models_dir = pathlib.Path(models_dir).as_posix()

                fold_relevance_dataset = self.get_kfold_relevance_dataset(args.kfold,
                                                                          args.include_testset_in_kfold,
                                                                          read_data_sets=False)
                fold_relevance_dataset.create_folds(fold_id, merged_data, relevance_dataset) 
                pipeline = self.create_pipeline_for_kfold(args)
                pipeline.run_pipeline(fold_relevance_dataset, fold_id)

            # removing intermediate directory and run kfold analysis
            self.local_io.rm_dir(os.path.join(self.data_dir_local, "tfrecord"))
            job_info = self.run_kfold_analysis(
                base_logs_dir, base_run_id, num_folds, args.kfold_analysis_metrics)

        except Exception as e:
            self.logger.error(
                "!!! Error in running Kfold CV !!!\n{}".format(str(e)))
            traceback.print_exc()
            job_status = "_FAILURE"
            job_info = "{}\n{}".format(str(e), traceback.format_exc())

    def run_pipeline(self, relevance_dataset=None):
        """
        Run the pipeline to train, evaluate and save the model.

        Parameters
        ----------
        relevance_dataset: RelevanceDataset
            RelevanceDataset used for running the pipeline. If none, the relevance dataset will be created.

        Returns
        -------
        dict
            Experiment tracking dictionary with metrics and metadata for the run.
            Used for model selection and hyperparameter optimization

        Notes
        -----
        Also populates a experiment tracking dictionary containing
        the metadata, model architecture and metrics generated by the model
        """
        experiment_tracking_dict = dict()
        try:
            job_status = "_SUCCESS"
            job_info = ""
            train_metrics = dict()
            test_metrics = dict()

            # Build dataset
            if not relevance_dataset:
                relevance_dataset = self.get_relevance_dataset()
                self.logger.info("Relevance Dataset created")

            # Build model
            relevance_model = self.get_relevance_model()
            self.logger.info("Relevance Model created")

            if self.args.execution_mode in {
                ExecutionModeKey.TRAIN_INFERENCE_EVALUATE,
                ExecutionModeKey.TRAIN_EVALUATE,
                ExecutionModeKey.TRAIN_INFERENCE,
                ExecutionModeKey.TRAIN_ONLY,
            }:

                # Train
                train_metrics = relevance_model.fit(
                    dataset=relevance_dataset,
                    num_epochs=self.args.num_epochs,
                    models_dir=self.models_dir_local,
                    logs_dir=self.logs_dir_local,
                    logging_frequency=self.args.logging_frequency,
                    monitor_metric=self.args.monitor_metric,
                    monitor_mode=self.args.monitor_mode,
                    patience=self.args.early_stopping_patience,
                )

            if self.args.execution_mode in {
                ExecutionModeKey.TRAIN_INFERENCE_EVALUATE,
                ExecutionModeKey.TRAIN_EVALUATE,
                ExecutionModeKey.EVALUATE_ONLY,
                ExecutionModeKey.INFERENCE_EVALUATE,
                ExecutionModeKey.INFERENCE_EVALUATE_RESAVE,
                ExecutionModeKey.EVALUATE_RESAVE,
            }:

                # Evaluate
                _, _, test_metrics = relevance_model.evaluate(
                    test_dataset=relevance_dataset.test,
                    inference_signature=self.args.inference_signature,
                    logging_frequency=self.args.logging_frequency,
                    group_metrics_min_queries=self.args.group_metrics_min_queries,
                    logs_dir=self.logs_dir_local,
                    compute_intermediate_stats=self.args.compute_intermediate_stats,
                )

            if self.args.execution_mode in {
                ExecutionModeKey.TRAIN_INFERENCE_EVALUATE,
                ExecutionModeKey.TRAIN_INFERENCE,
                ExecutionModeKey.INFERENCE_EVALUATE,
                ExecutionModeKey.INFERENCE_ONLY,
                ExecutionModeKey.INFERENCE_EVALUATE_RESAVE,
                ExecutionModeKey.INFERENCE_RESAVE,
            }:

                # Predict relevance scores
                relevance_model.predict(
                    test_dataset=relevance_dataset.test,
                    inference_signature=self.args.inference_signature,
                    additional_features={},
                    logs_dir=self.logs_dir_local,
                    logging_frequency=self.args.logging_frequency,
                )

            # Write experiment details to experiment tracking dictionary
            # Add command line script arguments
            experiment_tracking_dict.update(vars(self.args))

            # Add feature config information
            experiment_tracking_dict.update(
                self.feature_config.get_hyperparameter_dict())

            # Add train and test metrics
            experiment_tracking_dict.update(train_metrics)
            experiment_tracking_dict.update(test_metrics)

            # Add optimizer and lr schedule
            experiment_tracking_dict.update(
                relevance_model.model.optimizer.get_config())

            # Save model
            # NOTE: Model will be saved with the latest serving signatures
            if self.args.execution_mode in {
                ExecutionModeKey.TRAIN_INFERENCE_EVALUATE,
                ExecutionModeKey.TRAIN_EVALUATE,
                ExecutionModeKey.TRAIN_INFERENCE,
                ExecutionModeKey.TRAIN_ONLY,
                ExecutionModeKey.INFERENCE_EVALUATE_RESAVE,
                ExecutionModeKey.EVALUATE_RESAVE,
                ExecutionModeKey.INFERENCE_RESAVE,
                ExecutionModeKey.RESAVE_ONLY,
            }:
                # Save model
                relevance_model.save(
                    models_dir=self.models_dir_local,
                    preprocessing_keys_to_fns={},
                    postprocessing_fn=None,
                    required_fields_only=not self.args.use_all_fields_at_inference,
                    pad_sequence=self.args.pad_sequence_at_inference,
                    dataset=relevance_dataset,
                    experiment_details=experiment_tracking_dict
                )


            # temperature scaling
            if self.args.execution_mode in {
                ExecutionModeKey.TRAIN_INFERENCE_EVALUATE,
                ExecutionModeKey.TRAIN_EVALUATE,
                ExecutionModeKey.TRAIN_INFERENCE,
                ExecutionModeKey.TRAIN_ONLY,
                ExecutionModeKey.INFERENCE_EVALUATE_RESAVE,
                ExecutionModeKey.EVALUATE_RESAVE,
                ExecutionModeKey.INFERENCE_RESAVE,
            }:
                if CalibrationKey.CALIBRATION in self.model_config:
                    if self.model_config[CalibrationKey.CALIBRATION]['key'] == \
                            CalibrationKey.TEMPERATURE_SCALING:
                        kwargs = self.model_config[CalibrationKey.CALIBRATION][
                            CalibrationKey.ARGS] if CalibrationKey.ARGS in self.model_config[
                            CalibrationKey.CALIBRATION] else {}

                        results = relevance_model.calibrate(relevance_dataset=relevance_dataset,
                                                            logger=self.logger,
                                                            logs_dir_local=self.logs_dir_local,
                                                            **kwargs)

                        experiment_tracking_dict.update({CalibrationKey.TEMPERATURE:
                                                         results.position[0]})
                        # replacing the existing keras functional API model with the model with
                        # temperature scaling layer
                        relevance_model.add_temperature_layer(results.position[0])
                        # saving calibrated (with temperature scaling layer) model
                        relevance_model.save(
                            models_dir=self.models_dir_local,
                            preprocessing_keys_to_fns={},
                            postprocessing_fn=None,
                            required_fields_only=not self.args.use_all_fields_at_inference,
                            pad_sequence=self.args.pad_sequence_at_inference,
                            sub_dir="final_calibrated",
                            dataset=relevance_dataset,
                            experiment_details=experiment_tracking_dict
                        )

            job_info = pd.DataFrame.from_dict(
                experiment_tracking_dict, orient="index", columns=["value"]
            ).to_csv()

        except Exception as e:
            self.logger.error(
                "!!! Error Training Model: !!!\n{}".format(str(e)))
            traceback.print_exc()
            job_status = "_FAILURE"
            job_info = "{}\n{}".format(str(e), traceback.format_exc())

        # Finish
        self.finish(job_status, job_info)

        return experiment_tracking_dict

    def pre_processing_step(self):
        """
        Performs arbitrary pre-processing steps such as copying or transforming data that the rest of the code can not
        accommodate. It serves as a placeholder without an explicit implementation (returns self) in the base pipeline.
        We expect that users can extend it in their custom pipelines.
        """
        return self

    def post_training_step(self):
        """
        Performs arbitrary post-training steps such as copying or transforming data that the rest of the code can not
        accommodate. It serves as a placeholder without an explicit implementation (returns self) in the base pipeline.
        We expect that users can extend it in their custom pipelines.
        """
        return self

    def finish(self, job_status, job_info):
        """
        Wrap up the model training pipeline.
        Performs the following actions
            - save a job status file as _SUCCESS or _FAILURE to indicate job status.
            - delete temp data and models directories
            - if using spark IO, transfers models and logs directories to HDFS location from local directories
            - log overall run time of ml4ir job

        Parameters
        ----------
        job_status : str
            Tuple with first element _SUCCESS or _FAILURE
            second element
        job_info : str
            for _SUCCESS, is experiment tracking metrics and metadata
            for _FAILURE, is stacktrace of failure
        """
        # Write job status to file
        with open(os.path.join(self.logs_dir_local, job_status), "w") as f:
            f.write(job_info)

        # Delete temp data directories
        if self.data_format == DataFormatKey.CSV and self.args.kfold <= 1:
            self.local_io.rm_dir(os.path.join(self.data_dir_local, "tfrecord"))
        self.local_io.rm_dir(DefaultDirectoryKey.TEMP_DATA)
        self.local_io.rm_dir(DefaultDirectoryKey.TEMP_MODELS)

        if self.args.file_handler == FileHandlerKey.SPARK:
            # Copy logs and models to HDFS
            self.file_io.copy_to_hdfs(
                self.models_dir_local, self.models_dir, overwrite=True)
            self.file_io.copy_to_hdfs(
                self.logs_dir_local, self.logs_dir, overwrite=True)

        self.logger.info("Running post-training step.")
        self.post_training_step()
        self.logger.info("Post-training step done.")

        e = int(time.time() - self.start_time)
        self.logger.info(
            "Done! Elapsed time: {:02d}:{:02d}:{:02d}".format(
                e // 3600, (e % 3600 // 60), e % 60)
        )

        return self


def main(argv):
    # Define args
    args = get_args(argv)

    # Initialize Relevance Pipeline and run in train/inference mode
    rp = RelevancePipeline(args=args)
    rp.run()


if __name__ == "__main__":
    main(sys.argv[1:])



import tensorflow as tf
from tensorflow.keras import metrics
from tensorflow.python.ops import math_ops
import numpy as np
from tensorflow import Tensor
from tensorflow import dtypes

from ml4ir.base.model.metrics.metrics_impl import MetricState
from ml4ir.base.features.feature_config import FeatureConfig

from typing import Optional, Dict


class MeanMetricWrapper(metrics.Mean):
    """
    Class that wraps a stateless metric function with the Mean metric.

    Notes
    -----
    Original tensorflow implementation ->
    https://github.com/tensorflow/tensorflow/blob/r2.0/tensorflow/python/keras/metrics.py#L541-L590

    MeanMetricWrapper is not a public Class on tf.keras.metrics
    """

    def __init__(self, fn, name=None, dtype=None, **kwargs):
        """
        Create a `MeanMetricWrapper` instance

        Parameters
        ----------
        fn : function
            The metric function to wrap, with signature
            `fn(y_true, y_pred, **kwargs)`.
        name : str, optional
            string name of the metric instance.
        dtype : str, optional
            data type of the metric result.
        **kwargs : dict
            The keyword arguments that are passed on to `fn`.
        """
        super(MeanMetricWrapper, self).__init__(name=name, dtype=dtype)
        self._fn = fn
        self._fn_kwargs = kwargs

    def update_state(self, y_true, y_pred, sample_weight=None):
        """
        Accumulates metric statistics by computing the mean of the
        metric function

        Parameters
        ----------
        y_true : Tensor object
            The ground truth values. Shape : [batch_size, max_sequence_size, 1]
        y_pred : Tensor object
            The predicted values. Shape : [batch_size, max_sequence_size, 1]
        sample_weight : Tensor object
            Optional weighting of each example. Defaults to 1. Can be
            a `Tensor` whose rank is either 0, or the same rank as `y_true`,
            and must be broadcastable to `y_true`.

        Returns
        -------
            Update state of the metric

        Notes
        -----
        `y_true` and `y_pred` should have the same shape.
        """
        query_scores: Tensor = self._fn(y_true, y_pred, **self._fn_kwargs)
        return super(MeanMetricWrapper, self).update_state(
            query_scores, sample_weight=sample_weight
        )


class MeanRankMetric(MeanMetricWrapper):
    def __init__(
        self,
        feature_config: FeatureConfig,
        metadata_features: Dict,
        state: str = MetricState.NEW,
        name="MeanRankMetric",
        dtype: Optional[dtypes.DType] = None,
        **kwargs
    ):
        """
        Creates a `MeanRankMetric` instance to compute mean of rank

        Parameters
        ----------
        name : str
            string name of the metric instance.
        dtype : str, optional
            data type of the metric result.
        rank : Tensor object
            2D tensor representing ranks/rankitions of records in a query
        mask : Tensor object
            2D tensor representing 0/1 mask for padded records

        Notes
        -----
        rank and mask should be same shape as y_pred and y_true

        This metric creates two local variables, `total` and `count` that are used to
        compute the frequency with which `y_pred` matches `y_true`. This frequency is
        ultimately returned as `categorical accuracy`: an idempotent operation that
        simply divides `total` by `count`.
        `y_pred` and `y_true` should be passed in as vectors of probabilities, rather
        than as labels. If necessary, use `tf.one_hot` to expand `y_true` as a vector.
        If `sample_weight` is `None`, weights default to 1.
        Use `sample_weight` of 0 to mask values.
        """
        name = "{}_{}".format(state, name)
        # TODO: Handle Example dataset without mask and rank fields
        rank = metadata_features[feature_config.get_rank("node_name")]
        mask = metadata_features[feature_config.get_mask("node_name")]

        super(MeanRankMetric, self).__init__(
            self._compute, name, dtype=dtype, rank=rank, mask=mask
        )
        self.state = state

    def _compute(self, y_true, y_pred, rank, mask):
        """
        Compute mean rank metric

        Parameters
        ----------
        y_true : Tensor object
            Tensor object that contains the true label values
        y_pred : Tensor object
            Tensor object containing the predicted scores
        rank : Tensor object
            Tensor object that contains the rank of each record for a query
        masks : Tensor object
            Tensor object that contains 0/1 flag to identify which
            records were padded and thus should be excluded from metric computation
        """
        if self.state == "new":
            """Rerank using trained model"""
            # Convert y_pred for the masked records to -inf
            y_pred = tf.where(tf.equal(mask, 0), tf.constant(-np.inf), y_pred)

            # Convert predicted ranking scores into ranks for each record per query
            y_pred_ranks = tf.add(
                tf.argsort(
                    tf.argsort(y_pred, axis=-1, direction="DESCENDING", stable=True), stable=True
                ),
                tf.constant(1),
            )

            # Fetch indices of clicked records from y_true
            y_true_clicks = tf.where(tf.equal(tf.cast(y_true, tf.int32), tf.constant(1)))

            # Compute rank of clicked record from predictions
            click_ranks = tf.gather_nd(y_pred_ranks, indices=y_true_clicks)

        else:
            """Compute mean rank metric for existing data"""
            # Fetch indices of clicked records from y_true
            y_true_clicks = tf.where(tf.equal(tf.cast(y_true, tf.int32), tf.constant(1)))

            # Compute rank of clicked record from predictions
            click_ranks = tf.gather_nd(rank, indices=y_true_clicks)

        return self._get_matches_hook(click_ranks)

    def _get_matches_hook(self, y_pred_click_ranks):
        raise NotImplementedError


class MRR(MeanRankMetric):
    """
    Custom metric class to compute the Mean Reciprocal Rank.

    Calculates the mean of the reciprocal ranks of the
    clicked records from a list of queries.

    Examples
    --------
    >>> `y_true` is [[0, 0, 1], [0, 1, 0]]
    >>> `y_pred` is [[0.1, 0.9, 0.8], [0.05, 0.95, 0]]
    >>> `mask` is [[1, 1, 1], [1, 1, 1]] and
    >>> `rank` is [[1, 3, 2], [3, 1, 2]]
    >>> then the MRR is 0.75
    """

    def __init__(
        self,
        feature_config: FeatureConfig,
        metadata_features: Dict,
        name="MRR",
        state=MetricState.NEW,
        **kwargs
    ):
        """
        Creates a `MRR` instance to compute mean of reciprocal rank

        Parameters
        ----------
        feature_config : FeatureConfig object
            FeatureConfig object that defines the configuration for each model
            feature
        metadata_features : dict
            Dictionary of metadata feature tensors that can be used to compute
            custom metrics
        name : str
            Name of the metric
        state : {"new", "old"}
            State of the metric
        """
        super(MRR, self).__init__(
            feature_config=feature_config,
            metadata_features=metadata_features,
            name=name,
            state=state,
            **kwargs
        )

    def _get_matches_hook(self, y_pred_click_ranks):
        """
        Return reciprocal click ranks for MRR

        Parameters
        ----------
        y_pred_click_ranks: Tensor object
            Tensor object containing the ranks of the clicked records for each query

        Returns
        -------
        Tensor object
            Reciprocal ranks tensor
        """
        return math_ops.reciprocal(tf.cast(y_pred_click_ranks, tf.float32))


class ACR(MeanRankMetric):
    """
    Custom metric class to compute the Average Click Rank.

    Calculates the mean of the ranks of the
    clicked records from a list of queries.

    Examples
    --------
    >>> `y_true` is [[0, 0, 1], [0, 1, 0]]
    >>> `y_pred` is [[0.1, 0.9, 0.8], [0.05, 0.95, 0]]
    >>> `mask` is [[1, 1, 1], [1, 1, 1]] and
    >>> `rank` is [[1, 3, 2], [3, 1, 2]]
    >>> then the ACR is 1.50
    """

    def __init__(
        self,
        feature_config: FeatureConfig,
        metadata_features: Dict,
        name="ACR",
        state=MetricState.NEW,
        **kwargs
    ):
        """
        Creates a `ACR` instance to compute mean of rank

        Parameters
        ----------
        feature_config : FeatureConfig object
            FeatureConfig object that defines the configuration for each model
            feature
        metadata_features : dict
            Dictionary of metadata feature tensors that can be used to compute
            custom metrics
        name : str
            Name of the metric
        state : {"new", "old"}
            State of the metric
        """
        super(ACR, self).__init__(
            feature_config=feature_config,
            metadata_features=metadata_features,
            name=name,
            state=state,
            **kwargs
        )

    def _get_matches_hook(self, y_pred_click_ranks):
        """
        Return click ranks for ACR

        Parameters
        ----------
        y_pred_click_ranks: Tensor object
            Tensor object containing the ranks of the clicked records for each query

        Returns
        -------
        Tensor object
            Ranks tensor cast to float
        """
        return tf.cast(y_pred_click_ranks, tf.float32)



from ml4ir.applications.ranking.config.keys import LossTypeKey
from ml4ir.base.model.losses.loss_base import RelevanceLossBase


class RankingLossBase(RelevanceLossBase):
    """
    Abstract class that defines a Ranking loss function
    """

    def __init__(self, loss_type, loss_key, scoring_type):
        """
        Instantiate a RankingLossBase object

        Parameters
        ----------
        loss_type : str
            Type of the loss function - pointwise, pairwise, listwise
        loss_key : str
            Name of the loss function used
        scoring_type : str
            Type of scoring function - pointwise, pairwise, groupwise
        """
        self.loss_type = None
        self.loss_key = loss_key
        self.scoring_type = scoring_type


class PointwiseLossBase(RankingLossBase):
    """
    Abstract class that defines a pointwise ranking loss function
    """

    def __init__(self, loss_key="pointwise", scoring_type=""):
        """
        Instantiate a PointwiseLossBase object

        Parameters
        ----------
        loss_key : str
            Name of the loss function used
        scoring_type : str
            Type of scoring function - pointwise, pairwise, groupwise
        """
        super(PointwiseLossBase, self).__init__(
            loss_type=LossTypeKey.POINTWISE, loss_key=loss_key, scoring_type=scoring_type
        )


class PairwiseLossBase(RankingLossBase):
    """
    Abstract class that defines a pairwise ranking loss function
    """

    def __init__(self, loss_key="pairwise", scoring_type=""):
        """
        Instantiate a PairwiseLossBase object

        Parameters
        ----------
        loss_key : str
            Name of the loss function used
        scoring_type : str
            Type of scoring function - pointwise, pairwise, groupwise
        """
        super(PairwiseLossBase, self).__init__(
            loss_type=LossTypeKey.PAIRWISE, loss_key=loss_key, scoring_type=scoring_type
        )


class ListwiseLossBase(RankingLossBase):
    """
    Abstract class that defines a listwise ranking loss function
    """

    def __init__(self, loss_key="listwise", scoring_type=""):
        """
        Instantiate a ListwiseLossBase object

        Parameters
        ----------
        loss_key : str
            Name of the loss function used
        scoring_type : str
            Type of scoring function - pointwise, pairwise, groupwise
        """
        super(ListwiseLossBase, self).__init__(
            loss_type=LossTypeKey.LISTWISE, loss_key=loss_key, scoring_type=scoring_type
        )



import json
import yaml
import pandas as pd
from pyspark.sql import SparkSession
from logging import Logger

from ml4ir.base.io.file_io import FileIO

from typing import Optional


class SparkIO(FileIO):
    """Class defining the file I/O handler methods for the HDFS file system using spark"""

    def __init__(self, logger: Optional[Logger] = None):
        """
        Constructor method to create a FileIO handler object and
        set up spark session and hadoop file system handlers

        Parameters
        ----------
        logger : `Logger` object, optional
            logging handler object to instantiate FileIO object
            with the ability to log progress updates
        """
        self.logger = logger
        self.spark_session = SparkSession.builder.appName("ml4ir").getOrCreate()
        self.spark_context = self.spark_session.sparkContext
        self.hadoop_config = self.spark_context._jsc.hadoopConfiguration()
        self.hdfs = self.spark_context._gateway.jvm.org.apache.hadoop.fs.FileSystem.get(
            self.hadoop_config
        )
        self.local_fs = self.hdfs.getLocal(self.hadoop_config)

    def get_path_from_str(self, file_path: str):
        """
        Get Path object from string

        Parameters
        ----------
        file_path : str
            string file path

        Returns
        -------
        hadoop path
            Hadoop Path object
        """
        return self.spark_context._gateway.jvm.org.apache.hadoop.fs.Path(file_path)

    def read_df(
        self, infile: str, sep: str = ",", index_col: int = None
    ) -> Optional[pd.DataFrame]:
        """
        Load a pandas dataframe from a file

        Parameters
        ----------
        infile : str
            path to the csv input file; can be hdfs path
        sep : str, optional
            separator to use for loading file
        index_col : int, optional
            column to be used as index

        Returns
        -------
        `pandas.DataFrame`
            pandas dataframe loaded from file
        """
        self.log("Reading dataframe from : {}".format(infile))
        return (
            self.spark_session.read.format("csv")
            .option("header", "true")
            .option("inferschema", "true")
            .option("mode", "DROPMALFORMED")
            .option("mergeSchema", "true")
            .load(infile)
            .toPandas()
        )

    def read_df_list(self, infiles, sep=",", index_col=None) -> pd.DataFrame:
        """
        Load a pandas dataframe from a list of files

        Parameters
        ----------
        infiles : list of str
            paths to the csv input files; can be hdfs paths
        sep : str, optional
            separator to use for loading file
        index_col : int, optional
            column to be used as index

        Returns
        -------
        `pandas.DataFrame`
            pandas dataframe loaded from list of files

        Notes
        -----
        `sep` and `index_col` are not used in SparkIO
        """
        self.log("Reading {} files from [{}, ..".format(len(infiles), infiles[0]))
        return (
            self.spark_session.read.format("csv")
            .option("header", "true")
            .option("inferschema", "true")
            .option("mode", "DROPMALFORMED")
            .option("mergeSchema", "true")
            .load(infiles)
            .toPandas()
        )

    def read_text_file(self, infile) -> str:
        """
        Read text file and return as string

        Parameters
        ----------
        infile : str
            path to the text file

        Returns
        -------
        str
            file contents as a string
        """
        return "\n".join(
            [row["value"] for row in self.spark_session.read.format("text").load(infile).collect()]
        )

    def read_json(self, infile) -> dict:
        """
        Read JSON file and return a python dictionary

        Parameters
        ----------
        infile : str
            path to the json file; can be hdfs path

        Returns
        -------
        dict
            python dictionary loaded from file
        """
        self.log("Reading JSON file : {}".format(infile))
        return json.loads(self.read_text_file(infile))

    def read_yaml(self, infile) -> dict:
        """
        Read YAML file and return a python dictionary

        Parameters
        ----------
        infile : str
            path to the json file; can be hdfs path

        Returns
        -------
        dict
            python dictionary loaded from file
        """
        self.log("Reading YAML file : {}".format(infile))
        return yaml.safe_load(self.read_text_file(infile))

    def path_exists(self, path: str) -> bool:
        """
        Check if a path exists

        Parameters
        ----------
        path : str
            check if path exists

        Returns
        -------
        bool
            True if path exists; False otherwise
        """
        if self.hdfs.exists(self.get_path_from_str(path)):
            self.log("Path exists : {}".format(path))
            return True
        else:
            self.log("Path does not exist : {}".format(path))
            return False

    def rm_dir(self, dir_path: str):
        """
        Delete existing directory

        Parameters
        ----------
        dir_path : str
            path to directory to be removed
        """
        if self.path_exists(dir_path):
            self.hdfs.delete(self.get_path_from_str(dir_path), True)
            self.log("Directory deleted : {}".format(dir_path))

    def rm_file(self, file_path: str):
        """
        Deletes existing file_path

        Parameters
        ----------
        file_path : str
            path to file to be removed
        """
        if self.path_exists(file_path):
            self.hdfs.delete(self.get_path_from_str(file_path), True)
            self.log("File deleted : {}".format(file_path))

    def copy_from_hdfs(self, src: str, dest: str):
        """
        Copy a directory/file from HDFS to local filesystem

        Parameters
        ----------
        src : str
            String path to source(on HDFS)
        dest : str
            String path to destination(on local file system)
        """
        self.hdfs.copyToLocalFile(self.get_path_from_str(src), self.get_path_from_str(dest))

        self.log("Finished copying files from {} to {}".format(src, dest))

    def copy_to_hdfs(self, src: str, dest: str, overwrite=True):
        """
        Copy a directory/file to HDFS from local filesystem

        Parameters
        src : str
            String path to source(on local file system)
        dest : str
            String path to destination(on HDFS)
        overwrite : bool, optional
            Boolean to specify whether existing destination files should be overwritten
        """
        if overwrite:
            self.rm_dir(dest)

        self.hdfs.copyFromLocalFile(self.get_path_from_str(src), self.get_path_from_str(dest))

        self.log("Finished copying files from {} to {}".format(src, dest))



from ml4ir.base.config.keys import Key


class LossKey(Key):
    """Model loss keys that can be used with classification model"""

    CATEGORICAL_CROSS_ENTROPY = "categorical_cross_entropy"
    SIGMOID_CROSS_ENTROPY = "sigmoid_cross_entropy"


class MetricKey(Key):
    """Model metric keys that can be used with classification model"""

    CATEGORICAL_ACCURACY = "categorical_accuracy"
    TOP_5_CATEGORICAL_ACCURACY = "top_5_categorical_accuracy"



from tensorflow import saved_model


class Key:
    """Base class for Keys"""

    @classmethod
    def get_all_keys(cls):
        keys = list()
        for attr in dir(cls):
            if not callable(getattr(cls, attr)):
                if not attr.startswith("__"):
                    keys.append(cls.__dict__[attr])

        return keys


class ArchitectureKey(Key):
    """Model architecture keys"""

    DNN = "dnn"
    LINEAR = "linear"
    RNN = "rnn"


class OptimizerKey(Key):
    """Model optimizer keys"""

    ADAM = "adam"
    ADAGRAD = "adagrad"
    NADAM = "nadam"
    SGD = "sgd"
    RMS_PROP = "rms_prop"

class LearningRateScheduleKey(Key):
    """Learning rate schedule keys"""

    EXPONENTIAL = "exponential"
    CYCLIC = "cyclic"
    CONSTANT = 'constant'
    REDUCE_LR_ON_PLATEAU = 'reduce_lr_on_plateau'

class CyclicLearningRateType(Key):
    """Cyclic learning rate schedule type keys"""

    TRIANGULAR = "triangular"
    TRIANGULAR2 = "triangular2"
    EXPONENTIAL = "exponential"


class DataFormatKey(Key):
    """Data Format keys"""

    CSV = "csv"
    TFRECORD = "tfrecord"
    RANKLIB = "ranklib"


class DataSplitKey(Key):
    """Data Split keys"""

    TRAIN = "train"
    VALIDATION = "validation"
    TEST = "test"


class FeatureTypeKey(Key):
    """Feature Data Type keys"""

    NUMERIC = "numeric"
    STRING = "string"
    CATEGORICAL = "categorical"
    LABEL = "label"


class TFRecordTypeKey(Key):
    """Example or SequenceExample"""

    EXAMPLE = "example"
    SEQUENCE_EXAMPLE = "sequence_example"


class SequenceExampleTypeKey(Key):
    """Type of data in TFRecord SequenceExample protobuf"""

    SEQUENCE = "sequence"
    CONTEXT = "context"


class ExecutionModeKey(Key):
    """Type of execution mode for the pipeline"""

    TRAIN_EVALUATE = "train_evaluate"
    TRAIN_ONLY = "train_only"
    INFERENCE_ONLY = "inference_only"
    EVALUATE_ONLY = "evaluate_only"
    INFERENCE_EVALUATE = "inference_evaluate"
    TRAIN_INFERENCE_EVALUATE = "train_inference_evaluate"
    TRAIN_INFERENCE = "train_inference"
    INFERENCE_RESAVE = "inference_resave"
    EVALUATE_RESAVE = "evaluate_resave"
    INFERENCE_EVALUATE_RESAVE = "inference_evaluate_resave"
    RESAVE_ONLY = "resave_only"


class ServingSignatureKey(Key):
    """Serving signature names"""

    DEFAULT = saved_model.DEFAULT_SERVING_SIGNATURE_DEF_KEY
    TFRECORD = "serving_tfrecord"


class EncodingTypeKey(Key):
    """Types of embeddings"""

    BILSTM = "bilstm"
    CNN = "cnn"


class DefaultDirectoryKey(Key):
    """Default directory paths"""

    MODELS = "models"
    LOGS = "logs"
    DATA = "data"
    TEMP_DATA = "data/temp"
    TEMP_MODELS = "models/temp"


class FileHandlerKey(Key):
    """File handler type"""

    LOCAL = "local"
    SPARK = "spark"


class CalibrationKey(Key):
    CALIBRATION = "calibration"
    TEMPERATURE_SCALING = "temperature_scaling"
    TEMPERATURE = "temperature"
    ARGS = "args"


class PipelineType(Key):
    RANKING_PIPELINE = "RankingPipeline"
    CLASSIFICATION_PIPELINE = "ClassificationPipeline"



import tensorflow as tf
from tensorflow.keras import metrics

from ml4ir.base.model.metrics.metrics_impl import MetricState
from ml4ir.base.features.feature_config import FeatureConfig

from typing import Optional, Dict


class CategoricalAccuracy(metrics.CategoricalAccuracy):
    """
    Custom metric class to compute the Categorical Accuracy.

    Currently just a wrapper around tf.keras.metrics.CategoricalAccuracy
    to maintain consistency of arguments to __init__
    """

    def __init__(
        self,
        feature_config: FeatureConfig,
        metadata_features: Dict,
        name="categorical_accuracy",
        state=MetricState.NEW,
        **kwargs
    ):
        """
        Creates a CategoricalAccuracy instance

        Parameters
        ----------
        feature_config : FeatureConfig object
            FeatureConfig object that defines the configuration for each model
            feature
        metadata_features : dict
            Dictionary of metadata feature tensors that can be used to compute
            custom metrics
        name : str
            Name of the metric
        state : {"new", "old"}
            State of the metric
        """
        super(CategoricalAccuracy, self).__init__(name=name)


class Top5CategoricalAccuracy(metrics.TopKCategoricalAccuracy):
    """
    Custom metric class to compute the Top K Categorical Accuracy.

    Currently a wrapper around tf.keras.metrics.TopKCategoricalAccuracy that
    squeezes one dimension.
    It maintains consistency of arguments to __init__
    """

    def __init__(
        self,
        feature_config: Optional[FeatureConfig] = None,
        metadata_features: Dict = {},
        name="top_5_categorical_accuracy",
        state=MetricState.NEW,
        **kwargs
    ):
        """
        Creates a CategoricalAccuracy instance

        Parameters
        ----------
        feature_config : FeatureConfig object
            FeatureConfig object that defines the configuration for each model
            feature
        metadata_features : dict
            Dictionary of metadata feature tensors that can be used to compute
            custom metrics
        name : str
            Name of the metric
        state : {"new", "old"}
            State of the metric
        """
        super(Top5CategoricalAccuracy, self).__init__(name=name, k=5)

    def update_state(self, y_true, y_pred, sample_weight=None):
        """
        Squeeze the second dimension(axis=1) and compute top K categorical accuracy

        Parameters
        ----------
        y_true : Tensor object
            Tensor containing true class labels
            Shape : [batch_size, 1, num_classes]
        y_pred : Tensor object
            Tensor containing predicted scores for the classes
            Shape : [batch_size, 1, num_classes]
        sample_weight : dict
            Dictionary containing weights for the classes to measure weighted average metric

        Returns
        -------
        Tensor object
            Top K categorical accuracy computed on y_true and y_pred

        Notes
        -----
        Input shape is a 3 dimensional tensor of size
        (batch_size, 1, num_classes). We are squeezing
        the second dimension to follow the API of tf.keras.metrics.TopKCategoricalAccuracy

        Axis 1 of y_true and y_pred must be of size 1, otherwise `tf.squeeze`
        will throw error.
        """
        return super(Top5CategoricalAccuracy, self).update_state(
            tf.squeeze(y_true), tf.squeeze(y_pred), sample_weight=sample_weight
        )



import tensorflow as tf
from tensorflow.keras import layers
from tensorflow import feature_column
from tensorflow import lookup

import copy

from ml4ir.base.features.feature_fns.sequence import get_bilstm_encoding
from ml4ir.base.io.file_io import FileIO


CATEGORICAL_VARIABLE = "categorical_variable"


def categorical_embedding_with_hash_buckets(feature_tensor, feature_info, file_io: FileIO):
    """
    Converts a string feature tensor into a categorical embedding.
    Works by first converting the string into num_hash_buckets buckets
    each of size hash_bucket_size, then converting each hash bucket into
    a categorical embdding of dimension embedding_size. Finally, these embeddings
    are combined either through mean, sum or concat operations to generate the final
    embedding based on the feature_info.

    Parameters
    ----------
    feature_tensor : Tensor
        String feature tensor
    feature_info : dict
        Dictionary representing the configuration parameters for the specific feature from the FeatureConfig
    file_io : FileIO object
        FileIO handler object for reading and writing

    Returns
    -------
    Tensor object
        categorical embedding for the input feature_tensor

    Notes
    -----
    Args under feature_layer_info:
        num_hash_buckets : int
            number of different hash buckets to convert the input string into
        hash_bucket_size : int
            the size of each hash bucket
        embedding_size : int
            dimension size of the categorical embedding
        merge_mode : str
            can be one of "mean", "sum", "concat" representing the mode of combining embeddings from each categorical embedding
    """
    feature_layer_info = feature_info.get("feature_layer_info")
    embeddings_list = list()
    for i in range(feature_layer_info["args"]["num_hash_buckets"]):
        augmented_string = layers.Lambda(lambda x: tf.add(x, str(i)))(feature_tensor)

        hash_bucket = tf.strings.to_hash_bucket_fast(
            augmented_string, num_buckets=feature_layer_info["args"]["hash_bucket_size"]
        )
        embeddings_list.append(
            layers.Embedding(
                input_dim=feature_layer_info["args"]["hash_bucket_size"],
                output_dim=feature_layer_info["args"]["embedding_size"],
                name="categorical_embedding_{}_{}".format(feature_info.get("name"), i),
            )(hash_bucket)
        )

    embedding = None
    if feature_layer_info["args"]["merge_mode"] == "mean":
        embedding = tf.reduce_mean(
            embeddings_list,
            axis=0,
            name="categorical_embedding_{}".format(feature_info.get("name")),
        )
    elif feature_layer_info["args"]["merge_mode"] == "sum":
        embedding = tf.reduce_sum(
            embeddings_list,
            axis=0,
            name="categorical_embedding_{}".format(feature_info.get("name")),
        )
    elif feature_layer_info["args"]["merge_mode"] == "concat":
        embedding = tf.concat(
            embeddings_list,
            axis=-1,
            name="categorical_embedding_{}".format(feature_info.get("name")),
        )
    else:
        raise KeyError(
            "The merge_mode currently supported under categorical_embedding_with_hash_buckets are ['mean', 'sum', 'concat']. merge_mode specified in the feature config: {}".format(
                feature_layer_info["args"]["merge_mode"]
            )
        )

    embedding = tf.expand_dims(embedding, axis=1)

    return embedding


def categorical_embedding_with_indices(feature_tensor, feature_info, file_io: FileIO):
    """
    Converts input integer tensor into categorical embedding.
    Works by converting the categorical indices in the input feature_tensor,
    represented as integer values, into categorical embeddings based on the feature_info.

    Parameters
    ----------
    feature_tensor : Tensor object
        int feature tensor
    feature_info : dict
        Dictionary representing the configuration parameters for the specific feature from the FeatureConfig
    file_io : FileIO object
        FileIO handler object for reading and writing

    Returns
    -------
    Tensor object
        categorical embedding for the input feature_tensor

    Notes
    -----
    Args under feature_layer_info:
        num_buckets : int
            Maximum number of categorical values
        default_value : int
            default value to be assigned to indices out of the num_buckets range
        embedding_size : int
            dimension size of the categorical embedding

    String based categorical features should already be converted into numeric indices
    """
    feature_layer_info = feature_info.get("feature_layer_info")

    categorical_fc = feature_column.categorical_column_with_identity(
        CATEGORICAL_VARIABLE,
        num_buckets=feature_layer_info["args"]["num_buckets"],
        default_value=feature_layer_info["args"].get("default_value", None),
    )
    embedding_fc = feature_column.embedding_column(
        categorical_fc, dimension=feature_layer_info["args"]["embedding_size"], trainable=True
    )

    embedding = layers.DenseFeatures(
        embedding_fc,
        name="{}_embedding".format(feature_info.get("node_name", feature_info["name"])),
    )({CATEGORICAL_VARIABLE: feature_tensor})
    embedding = tf.expand_dims(embedding, axis=1)

    return embedding


def categorical_embedding_to_encoding_bilstm(feature_tensor, feature_info, file_io: FileIO):
    """
    Encode a string tensor into categorical embedding.
    Works by converting the string into a word sequence and then generating a categorical/char embedding for each words
    based on the List of strings that form the vocabulary set of categorical values, defined by the argument
    vocabulary_file.
    The char/byte embeddings are then combined using a biLSTM.

    Parameters
    ----------
    feature_tensor : Tensor object
        String feature tensor that is to be encoded
    feature_info : dict
        Dictionary representing the feature_config for the input feature
    file_io : FileIO object
        FileIO handler object for reading and writing

    Returns
    -------
    Tensor object
        Encoded feature tensor

    Notes
    -----
    Args under feature_layer_info:
        vocabulary_file : string
            path to vocabulary CSV file for the input tensor containing the vocabulary to look-up.
            uses the "key" named column as vocabulary of the 1st column if no "key" column present.
        max_length: int
            max number of rows to consider from the vocabulary file.
                        if null, considers the entire file vocabulary.
        embedding_size : int
            dimension size of the embedding;
                        if null, then the tensor is just converted to its one-hot representation
        encoding_size : int
            dimension size of the sequence encoding computed using a biLSTM

    The input dimension for the embedding is fixed to 256 because the string is
    converted into a bytes sequence.
    """
    args = feature_info.get("feature_layer_info")["args"]

    (
        categorical_indices,
        vocabulary_size,
        num_oov_buckets,
    ) = categorical_indices_from_vocabulary_file(feature_info, feature_tensor, file_io)

    input_dim = vocabulary_size + num_oov_buckets if num_oov_buckets else vocabulary_size
    categorical_embeddings = layers.Embedding(
        input_dim=input_dim,
        output_dim=args["embedding_size"],
        mask_zero=True,
        input_length=args.get("max_length"),
    )(categorical_indices)

    categorical_embeddings = tf.squeeze(categorical_embeddings, axis=1)
    kernel_initializer = args.get("lstm_kernel_initializer", "glorot_uniform")
    encoding = get_bilstm_encoding(
        embedding=categorical_embeddings,
        lstm_units=int(args["encoding_size"] / 2),
        kernel_initializer=kernel_initializer,
    )
    return encoding


def categorical_embedding_with_vocabulary_file(feature_tensor, feature_info, file_io: FileIO):
    """
    Converts a string tensor into a categorical embedding representation.
    Works by using a vocabulary file to convert the string tensor into categorical indices
    and then converting the categories into embeddings based on the feature_info.

    Parameters
    ----------
    feature_tensor : Tensor object
        String feature tensor
    feature_info : dict
        Dictionary representing the configuration parameters for the specific feature from the FeatureConfig
    file_io : FileIO object
        FileIO handler object for reading and writing

    Returns
    -------
    Tensor object
        Categorical embedding representation of input feature_tensor

    Notes
    -----
    Args under feature_layer_info:
        vocabulary_file : string
            path to vocabulary CSV file for the input tensor containing the vocabulary to look-up.
                        uses the "key" named column as vocabulary of the 1st column if no "key" column present.
        max_length : int
            max number of rows to consider from the vocabulary file.
                        if null, considers the entire file vocabulary.
        num_oov_buckets : int
            number of out of vocabulary buckets/slots to be used to
                         encode strings into categorical indices
        embedding_size : int
            dimension size of categorical embedding

    The vocabulary CSV file must contain two columns - key, id,
    where the key is mapped to one id thereby resulting in a
    many-to-one vocabulary mapping.
    If id field is absent, a unique whole number id is assigned by default
    resulting in a one-to-one mapping
    """
    (
        categorical_indices,
        vocabulary_size,
        num_oov_buckets,
    ) = categorical_indices_from_vocabulary_file(feature_info, feature_tensor, file_io)
    feature_info_new = copy.deepcopy(feature_info)
    feature_info_new["feature_layer_info"]["args"]["num_buckets"] = (
        vocabulary_size + num_oov_buckets
    )
    feature_info_new["feature_layer_info"]["args"]["default_value"] = vocabulary_size

    embedding = categorical_embedding_with_indices(
        feature_tensor=categorical_indices, feature_info=feature_info_new, file_io=file_io
    )

    return embedding


class CategoricalDropout(layers.Layer):
    """
    Custom Dropout class for categorical indices

    Examples
    --------
    >>> inputs: [[1, 2, 3], [4, 1, 2]]
    >>> dropout_rate = 0.5

    >>> When training, output: [[0, 0, 3], [0, 1, 2]]
    >>> When testing, output: [[1, 2, 3], [4, 1, 2]]

    Notes
    -----
    At training time, mask indices to 0 at dropout_rate

    This works similar to tf.keras.layers.Dropout without the scaling
    Ref: https://www.tensorflow.org/api_docs/python/tf/keras/layers/Dropout
    """

    def __init__(self, dropout_rate, seed=None, **kwargs):
        """
        Parameters
        ----------
        dropout_rate : float
            fraction of units to drop, i.e., set to OOV token 0
        seed : int
            random seed for sampling to mask/drop categorical labels

        Notes
        -----
        We define OOV index to be 0 for this function and when dropout is applied, it converts p% of the values to 0(which is the OOV index). This allows us to train a good average embedding for the OOV token.
        """
        super(CategoricalDropout, self).__init__(**kwargs)
        self.dropout_rate = dropout_rate
        self.seed = seed

    def get_config(self):
        """
        Get config for the CategoricalDropout tensorflow layer

        Returns
        -------
        dict
            Configuration dictionary for the tensorflow layer
        """
        config = super(CategoricalDropout, self).get_config()
        config.update({"dropout_rate": self.dropout_rate, "seed": self.seed})
        return config

    def call(self, inputs, training=None):
        """
        Run the CategoricalDropout layer by masking input labels to OOV
        index 0 at `dropout_rate`

        Parameters
        ----------
        input : Tensor object
            int categorical index tensor to be masked
        training : bool
            If the layer should be run as training or not

        Returns
        -------
        Tensor object
            Masked tensor object with values set to 0 at probability of dropout_rate
        """
        if training:
            return tf.math.multiply(
                tf.cast(
                    tf.random.uniform(shape=tf.shape(inputs), seed=self.seed) >= self.dropout_rate,
                    dtype=tf.int64,
                ),
                inputs,
            )
        else:
            return inputs


def categorical_embedding_with_vocabulary_file_and_dropout(
    feature_tensor, feature_info, file_io: FileIO
):
    """
    Converts a string tensor into a categorical embedding representation.
    Works by using a vocabulary file to convert the string tensor into categorical indices
    and then converting the categories into embeddings based on the feature_info.
    Also uses a dropout to convert categorical indices to the OOV index of 0 at a rate of dropout_rate

    Parameters
    ----------
    feature_tensor : Tensor object
        String feature tensor
    feature_info : dict
        Dictionary representing the configuration parameters for the specific feature from the FeatureConfig
    file_io : FileIO object
        FileIO handler object for reading and writing

    Returns
    -------
    Tensor object
        Categorical embedding representation of input feature_tensor

    Notes
    -----
    Args under feature_layer_info:
        vocabulary_file : str
            path to vocabulary CSV file for the input tensor
        dropout_rate : float
            rate at which to convert categorical indices to OOV
        embedding_size : int
            dimension size of categorical embedding

    The vocabulary CSV file must contain two columns - key, id,
    where the key is mapped to one id thereby resulting in a
    many-to-one vocabulary mapping.
    If id field is absent, a unique natural number id is assigned by default
    resulting in a one-to-one mapping

    OOV index will be set to 0
    num_oov_buckets will be 0
    """
    feature_layer_info = feature_info.get("feature_layer_info")
    (
        feature_tensor_indices,
        vocabulary_size,
        num_oov_buckets,
    ) = categorical_indices_from_vocabulary_file(feature_info, feature_tensor, file_io)

    # Mask indices to OOV key of 0 at dropout_rate
    feature_tensor_indices = CategoricalDropout(
        dropout_rate=feature_layer_info["args"]["dropout_rate"]
    )(feature_tensor_indices)

    feature_info_new = copy.deepcopy(feature_info)
    feature_info_new["feature_layer_info"]["args"]["num_buckets"] = vocabulary_size
    feature_info_new["feature_layer_info"]["args"]["default_value"] = 0

    embedding = categorical_embedding_with_indices(
        feature_tensor=feature_tensor_indices, feature_info=feature_info_new, file_io=file_io
    )

    return embedding


def categorical_indicator_with_vocabulary_file(feature_tensor, feature_info, file_io: FileIO):
    """
    Converts a string tensor into a categorical one-hot representation.
    Works by using a vocabulary file to convert the string tensor into categorical indices
    and then converting the categories into one-hot representation.

    Parameters
    ----------
    feature_tensor : Tensor object
        String feature tensor
    feature_info : dict
        Dictionary representing the configuration parameters for the specific feature from the FeatureConfig
    file_io : FileIO object
        FileIO handler object for reading and writing

    Returns
    -------
    Tensor object
        Categorical one-hot representation of input feature_tensor

    Notes
    -----
    Args under feature_layer_info:
        vocabulary_file : string
            path to vocabulary CSV file for the input tensor containing the vocabulary to look-up.
            uses the "key" named column as vocabulary of the 1st column if no "key" column present.
        max_length : int
            max number of rows to consider from the vocabulary file.
            if null, considers the entire file vocabulary.
        num_oov_buckets : int, optional
            number of out of vocabulary buckets/slots to be used to
            encode strings into categorical indices. If not specified, the default is 1.

    The vocabulary CSV file must contain two columns - key, id,
    where the key is mapped to one id thereby resulting in a
    many-to-one vocabulary mapping.
    If id field is absent, a unique whole number id is assigned by default
    resulting in a one-to-one mapping
    """
    #
    ##########################################################################
    #
    # NOTE:
    # Current bug[1] with saving a Keras model when using
    # feature_column.categorical_column_with_vocabulary_list.
    # Tracking the issue currently and should be able to upgrade
    # to current latest stable release 2.2.0 to test.
    #
    # Can not use TF2.1.0 due to issue[2] regarding saving Keras models with
    # custom loss, metric layers
    #
    # Can not use TF2.2.0 due to issues[3, 4] regarding incompatibility of
    # Keras Functional API models and Tensorflow
    #
    # References:
    # [1] https://github.com/tensorflow/tensorflow/issues/31686
    # [2] https://github.com/tensorflow/tensorflow/issues/36954
    # [3] https://github.com/tensorflow/probability/issues/519
    # [4] https://github.com/tensorflow/tensorflow/issues/35138
    #
    # CATEGORICAL_VARIABLE = "categorical_variable"
    # categorical_fc = feature_column.categorical_column_with_vocabulary_list(
    #     CATEGORICAL_VARIABLE,
    #     vocabulary_list=vocabulary_list,
    #     default_value=feature_layer_info["args"].get("default_value", -1),
    #     num_oov_buckets=feature_layer_info["args"].get("num_oov_buckets", 0),
    # )
    #
    # indicator_fc = feature_column.indicator_column(categorical_fc)
    #
    # categorical_one_hot = layers.DenseFeatures(
    #     indicator_fc,
    #     name="{}_one_hot".format(feature_info.get("node_name", feature_info["name"])),
    # )({CATEGORICAL_VARIABLE: feature_tensor})
    # categorical_one_hot = tf.expand_dims(categorical_one_hot, axis=1)
    #
    ##########################################################################
    #
    (
        feature_tensor_indices,
        vocabulary_size,
        num_oov_buckets,
    ) = categorical_indices_from_vocabulary_file(feature_info, feature_tensor, file_io)

    categorical_identity_fc = feature_column.categorical_column_with_identity(
        CATEGORICAL_VARIABLE, num_buckets=vocabulary_size + num_oov_buckets
    )
    indicator_fc = feature_column.indicator_column(categorical_identity_fc)

    categorical_one_hot = layers.DenseFeatures(
        indicator_fc,
        name="{}_one_hot".format(feature_info.get("node_name", feature_info["name"])),
    )({CATEGORICAL_VARIABLE: feature_tensor_indices})
    categorical_one_hot = tf.expand_dims(categorical_one_hot, axis=1)

    return categorical_one_hot


def categorical_indices_from_vocabulary_file(feature_info, feature_tensor, file_io):
    """
    Extract the vocabulary (encoding and values) from the stated vocabulary_file inside feature_info.
    And encode the feature_tensor with the vocabulary.

    Parameters
    ----------
    feature_tensor : Tensor object
        String feature tensor
    feature_info : dict
        Dictionary representing the configuration parameters for the specific feature from the FeatureConfig
    file_io : FileIO object
        FileIO handler object for reading and writing

    Returns
    -------
    categorical_indices : Tensor object
        corresponding to encode of the feature_tensor based on the vocabulary.
    num_oov_buckets : int
        applied num_oov_buckets
    vocabulary_keys : list
        values of the vocabulary stated in the vocabulary_file.
    """
    vocabulary_keys, vocabulary_ids = get_vocabulary_info(feature_info, file_io)
    args = feature_info.get("feature_layer_info")["args"]

    if "dropout_rate" in args:
        default_value = 0  # Default OOV index when using dropout
        if args.get("num_oov_buckets"):
            raise RuntimeError(
                "Cannot have both dropout_rate and num_oov_buckets set. "
                "OOV buckets are not supported with dropout"
            )
        num_oov_buckets = None
        vocabulary_size = len(set(vocabulary_keys)) + 1  # one more for default value
    else:
        default_value = None
        num_oov_buckets = args.get("num_oov_buckets", 1)
        vocabulary_size = len(set(vocabulary_keys))

    lookup_table = VocabLookup(
        vocabulary_keys=vocabulary_keys,
        vocabulary_ids=vocabulary_ids,
        num_oov_buckets=num_oov_buckets,
        default_value=default_value,
        feature_name=feature_info.get("node_name", feature_info["name"]),
    )
    categorical_indices = lookup_table(feature_tensor)

    return categorical_indices, vocabulary_size, num_oov_buckets


class VocabLookup(layers.Layer):
    """
    The class defines a keras layer wrapper around a tf lookup table using the given vocabulary list.
    Maps each entry of a vocabulary list into categorical indices.

    Attributes
    ----------
    vocabulary_list : list
        List of strings that form the vocabulary set of categorical values
    num_oov_buckets : int
        Number of buckets to be used for out of vocabulary strings
    default_value : int
        Default value to strbe used for OOV values
    feature_name : str
        Name of the input feature tensor
    lookup_table : LookupTable object
        Tensorflow look up table that maps strings to integer indices

    Notes
    -----
    Issue[1] with using LookupTable with keras symbolic tensors; expects eager tensors.

    Ref: https://github.com/tensorflow/tensorflow/issues/38305
    """

    def __init__(
        self,
        vocabulary_keys,
        vocabulary_ids,
        num_oov_buckets: int = None,
        default_value: int = None,
        feature_name="categorical_variable",
    ):
        super(VocabLookup, self).__init__(trainable=False, dtype=tf.int64)
        self.vocabulary_keys = vocabulary_keys
        self.vocabulary_ids = vocabulary_ids
        self.vocabulary_size = len(set(vocabulary_ids))
        self.num_oov_buckets = num_oov_buckets
        self.default_value = default_value
        self.feature_name = feature_name

    def build(self, input_shape):
        """
        Defines a Lookup Table  using a KeyValueTensorInitializer to map the keys to the IDs.
        Allows definition of two types of lookup tables based on whether the user specifies num_oov_buckets or the default_value
        """
        table_init = lookup.KeyValueTensorInitializer(
            keys=self.vocabulary_keys,
            values=self.vocabulary_ids,
            key_dtype=tf.string,
            value_dtype=tf.int64,
        )

        """
        NOTE:
        If num_oov_buckets are specified, use a StaticVocabularyTable
        otherwise, use a StaticHashTable with the specified default_value

        For most cases, StaticVocabularyTable is sufficient. But when we want to use a custom default_value(like in the case of the dropout function), we need to use StaticHashTable.
        """
        if self.num_oov_buckets is not None:
            self.lookup_table = lookup.StaticVocabularyTable(
                initializer=table_init,
                num_oov_buckets=self.num_oov_buckets,
                name="{}_lookup_table".format(self.feature_name),
            )
        elif self.default_value is not None:
            self.lookup_table = lookup.StaticHashTable(
                initializer=table_init,
                default_value=self.default_value,
                name="{}_lookup_table".format(self.feature_name),
            )
        else:
            raise KeyError("You must specify either num_oov_buckets or default_value")
        self.built = True

    def call(self, input_text):
        """
        Convert string tensors to numeric indices using lookup table

        Parameters
        ----------
        input_text : Tensor object
            String categorical tensor

        Returns
        -------
        Tensor object
            Numeric tensor object with corresponding lookup indices
        """
        return self.lookup_table.lookup(input_text)

    def get_config(self):
        """
        Get tensorflow configuration for the lookup table

        Returns
        -------
        dict
            Configuration dictionary for the lookup table layer
        """
        config = super(VocabLookup, self).get_config()
        config.update(
            {
                "vocabulary_keys": self.vocabulary_keys,
                "vocabulary_ids": self.vocabulary_ids,
                "vocabulary_size": self.vocabulary_size,
                "num_oov_buckets": self.num_oov_buckets,
                "feature_name": self.feature_name,
            }
        )
        return config


def get_vocabulary_info(feature_info, file_io):
    """
    Extract the vocabulary (encoding and values) from the stated vocabulary_file inside feature_info.

    Parameters
    ----------
    feature_info : dict
        Dictionary representing the configuration parameters for the specific feature from the FeatureConfig
    file_io : FileIO object
        FileIO handler object for reading and writing files

    Returns
    -------
    vocabulary_keys : list
        values of the vocabulary stated in the vocabulary_file.
    vocabulary_ids : list
        corresponding encoding ids (values of the vocabulary_keys).

    Notes
    -----
    Args under `feature_layer_info`
        vocabulary_file : str
            path to vocabulary CSV file for the input tensor containing the vocabulary to look-up.
            uses the "key" named column as vocabulary of the 1st column if no "key" column present.
        max_length : int
            max number of rows to consider from the vocabulary file.
            if null, considers the entire file vocabulary.
        default_value : int
            default stated value in the configure used to replace missing data points.
    """
    args = feature_info.get("feature_layer_info")["args"]
    vocabulary_df = file_io.read_df(args["vocabulary_file"])
    if "key" in vocabulary_df.columns:
        vocabulary_keys = vocabulary_df["key"]
    else:
        vocabulary_keys = vocabulary_df.iloc[:, 0]
    if "max_length" in args:
        vocabulary_keys = vocabulary_keys[: args["max_length"]]
    if "default_value" in feature_info:
        vocabulary_keys = vocabulary_keys.fillna(feature_info["default_value"])
    vocabulary_keys = vocabulary_keys.values
    if "dropout_rate" in args:
        # NOTE: If a dropout_rate is specified, then reserve 0 as the OOV index
        vocabulary_ids = (
            vocabulary_df["id"].values
            if "id" in vocabulary_df
            else list(range(1, len(vocabulary_keys) + 1))
        )
        if 0 in vocabulary_ids:
            raise ValueError(
                "Can not use ID 0 with dropout. Use categorical_embedding_with_vocabulary_file instead."
            )
    else:
        vocabulary_ids = (
            vocabulary_df["id"].values
            if "id" in vocabulary_df
            else list(range(len(vocabulary_keys)))
        )
    return vocabulary_keys, vocabulary_ids



import os
import json
import shutil
import pandas as pd
import numpy as np
import gzip
import sys
import csv
import glob
import yaml
import logging

from ml4ir.base.io.file_io import FileIO

from typing import Optional

from io import StringIO  # type: ignore


class LocalIO(FileIO):
    """Class defining the file I/O handler methods for the local file system"""

    def make_directory(self, dir_path: str, clear_dir: bool = False):
        """
        Create directory structure specified recursively

        Parameters
        ----------
        dir_path : str
            path for directory to be create
        clear_dir : bool, optional
            clear contents on existing directory
        """

        if os.path.exists(dir_path):
            if clear_dir:
                try:
                    shutil.rmtree(dir_path)
                except shutil.Error as e:  # Potentially OSError, IOError
                    raise Exception("Cannot remove local folder, with error: {}".format(e))
            else:
                return dir_path

        # Knowing that folder does not exist, create it from scratch
        os.makedirs(dir_path)

    def read_df(
        self, infile: str, sep: str = ",", index_col: int = None
    ) -> Optional[pd.DataFrame]:
        """
        Load a pandas dataframe from a file

        Parameters
        ----------
        infile : str
            path to the csv input file; can be hdfs path
        sep : str, optional
            separator to use for loading file
        index_col : int, optional
            column to be used as index

        Returns
        -------
        `pandas.DataFrame`
            pandas dataframe loaded from file
        """
        self.log("Loading dataframe from path : {}".format(infile))

        if infile.endswith(".gz"):
            fp = gzip.open(os.path.expanduser(infile), "rb")
        else:
            fp = open(os.path.expanduser(infile), "r")

        # Redirect stderr to custom string IO
        stderr_old = sys.stderr
        sys.stderr = bad_lines_io = StringIO()

        try:
            df: pd.DataFrame = pd.read_csv(
                fp,
                sep=sep,
                index_col=index_col,
                skipinitialspace=True,
                quotechar='"',
                escapechar="\\",
                error_bad_lines=False,
                warn_bad_lines=True,
                engine="c",
            )

        except Exception as e:
            self.log("Error while reading : {}\n{}".format(fp, e), mode=logging.WARN)
            return None
        # Get the bad line string value and close the string IO
        bad_lines = bad_lines_io.getvalue()
        bad_lines_io.close()
        sys.stderr = stderr_old

        # Log any bad lines
        if bad_lines:
            self.log("Bad lines were found in the file : {}\n{}".format(infile, bad_lines))

        fp.close()
        return df

    def read_df_list(self, infiles, sep=",", index_col=None) -> pd.DataFrame:
        """
        Load a pandas dataframe from a list of files

        Parameters
        ----------
        infiles : list of str
            paths to the csv input files; can be hdfs paths
        sep : str, optional
            separator to use for loading file
        index_col : int, optional
            column to be used as index

        Returns
        -------
        `pd.DataFrame`
            pandas dataframe loaded from file
        """
        self.log("Reading {} files from [{}, ..".format(len(infiles), infiles[0]))
        return pd.concat(
            [self.read_df(infile, sep=sep, index_col=index_col) for infile in infiles]
        )

    def write_df(self, df, outfile: str = None, sep: str = ",", index: bool = True) -> str:
        """
        Write a pandas dataframe to a file

        Parameters
        ----------
        df : `pandas.DataFrame`
            dataframe to be written
        outfile : str
            path to the csv output file; can NOT be hdfs path currently
        sep : str
            separator to use for loading file
        index : int
            boolean specifying if index should be saved

        Returns
        -------
        str
            dataframe in csv form if outfile is None
        """
        self.log("Writing dataframe to : {}".format(outfile))
        # Decode the binary stream in utf-8 to avoid the extra b'...' formating when writing csv predictions
        np.set_printoptions(formatter={'all': lambda x: str(x.decode('utf-8')) if isinstance(x, bytes) else str(x)},
                            linewidth=sys.maxsize, threshold=sys.maxsize)
        output = df.to_csv(
            sep=sep, index=index, quotechar='"', escapechar="\\", quoting=csv.QUOTE_NONNUMERIC
        )
        output = output.replace("\\", "\\\\")

        if outfile:
            fp = open(outfile, "w")
            fp.write(output)
            fp.close()
            return ""
        else:
            return output

    def read_json(self, infile) -> dict:
        """
        Read JSON file and return a python dictionary

        Parameters
        ----------
        infile : str
            path to the json file; can be hdfs path

        Returns
        -------
        dict
            python dictionary loaded from file
        """
        self.log("Reading JSON file from : {}".format(infile))
        return json.load(open(infile, "r"))

    def read_yaml(self, infile) -> dict:
        """
        Read YAML file and return a python dictionary

        Parameters
        ----------
        infile : str
            path to the json file; can be hdfs path

        Returns
        -------
        dict
            python dictionary loaded from file
        """
        self.log("Reading YAML file from : {}".format(infile))
        return yaml.safe_load(open(infile, "r"))

    def write_json(self, json_dict: dict, outfile: str):
        """
        Write dictionary to a JSON file

        Parameters
        ----------
        json_dict : dict
            dictionary to be dumped to json file
        outfile : str
            path to the output file
        """
        self.log("Writing JSON dictionary to : {}".format(outfile))
        json.dump(json_dict, open(outfile, "w"), indent=4, sort_keys=True)

    def path_exists(self, path: str) -> bool:
        """
        Check if a path exists

        Parameters
        ----------
        path : str
            check if path exists

        Returns
        -------
        bool
            True if path exists; False otherwise
        """
        if os.path.exists(path):
            self.log("Path exists: {}".format(path))
            return True
        else:
            self.log("Path does not exists: {}".format(path))
            return False

    def get_files_in_directory(self, indir: str, extension=".csv", prefix=""):
        """
        Get list of files in a directory

        Parameters
        ----------
        indir : str
            input directory to search for files
        extension : str, optional
            extension of the files to search for
        prefix : str, optional
            string file name prefix to narrow search

        Returns
        -------
        list of str
            list of file path strings
        """
        files_in_directory = sorted(
            glob.glob(os.path.join(indir, "{}*{}".format(prefix, extension)))
        )
        self.log("{} files found under {}".format(len(files_in_directory), indir))
        return files_in_directory

    def clear_dir_contents(self, dir_path: str):
        """
        Clear contents of existing directory

        Parameters
        ----------
        dir_path :  str
            path to directory to be cleared
        """
        for dir_content in glob.glob(os.path.join(dir_path, "*")):
            if os.path.isfile(dir_content):
                os.remove(dir_content)
            elif os.path.isdir(dir_content):
                shutil.rmtree(dir_content)
        self.log("Directory cleared : {}".format(dir_path))

    def rm_dir(self, dir_path: str):
        """
        Delete existing directory

        Parameters
        ----------
        dir_path : str
            path to directory to be removed
        """
        if os.path.isdir(dir_path):
            shutil.rmtree(dir_path)
            self.log("Directory deleted : {}".format(dir_path))

    def rm_file(self, file_path: str):
        """
        Delete existing file_path

        Parameters
        ----------
        file_path : str
            path to file to be removed
        """
        if os.path.isfile(file_path):
            os.remove(file_path)
            self.log("File deleted : {}".format(file_path))

    def save_numpy_array(self, np_array, file_path: str, allow_pickle=True, zip=True, **kwargs):
        """
        Save a numpy array to disk

        Parameters
        ----------
        np_array : numpy array or list of numpy arrays
            Array like numpy object to be saved
        file_path : str
            file path to save the object to
        allow_pickle : bool, optional
            Allow pickling of objects while saving
        zip : bool, optional,
            use np.savez to save the numpy arrays, allows passing in python list

        Notes
        -----
        Used to save individual model layer weights for transfer learning.

        If using zip=True, the np_array has to be a python list
        tensorflow layer weights are lists of arrays.
        np.save() can not be used for saving list of numpy arrays directly
        as it tries to manually convert the list into a numpy array, leading
        to errors with numpy shape.
        savez allows us to save each list item in separate files and abstracts this step for end user.
        """
        if zip:
            """
            NOTE: In this case, the np_array has to be a python list

            tensorflow layer weights are lists of arrays.
            np.save() can not be used for saving list of numpy arrays directly
            as it tries to manually convert the list into a numpy array, leading
            to errors with numpy shape.
            savez allows us to save each list item in separate files and abstracts this step for end user.
            """
            np.savez(file_path, *np_array)
        else:
            np.save(file_path, arr=np_array, allow_pickle=allow_pickle, **kwargs)

    def load_numpy_array(self, file_path, allow_pickle=True, unzip=True, **kwargs):
        """
        Load a numpy array from disk

        Parameters
        ----------
        file_path : str
            file path to load the numpy object from
        allow_pickle : bool, optional
            Allow pickling of objects while loading
        unzip : bool, optional
            To unzip the numpy array saved as a zip file. Used when saved with zip=True

        Returns
        -------
        list of numpy arrays
            python list of numpy arrays

        Notes
        -----
        Used to load individual model layer weights for transfer learning
        """
        np_array = np.load(file_path, allow_pickle=allow_pickle, **kwargs)

        if unzip:
            np_array_list = list()
            for np_file in np_array.files:
                np_array_list.append(np_array[np_file])
            return np_array_list
        else:
            return np_array



import tensorflow as tf

from ml4ir.base.io.file_io import FileIO


def tf_native_op(feature_tensor: tf.Tensor, feature_info: dict, file_io: FileIO):
    """
    Run a series of tensorflow native operations on the input feature tensor.
    The functions will be applied in the order they are specified.

    Parameters
    ----------
    feature_tensor : Tensor
        Input feature tensor
    feature_info : dict
        Dictionary representing the configuration parameters for the specific feature from the FeatureConfig
    file_io : FileIO object
        FileIO handler object for reading and writing

    Returns
    -------
    Tensor object
        Modified feature tensor after applying all the specified ops

    Notes
    -----
    Args under feature_layer_info:
        ops: list of dict
            List of function specifications with associated arguments
            
            Arguments under opts:
                fn : str
                    Tensorflow native function name. Should start with tf.
                    Example: tf.math.log or tf.clip_by_value
                args : dict
                    Keyword arguments to be passed to the tensorflow function
    """
    feature_node_name = feature_info.get("node_name", feature_info.get("name"))
    tf_ops = feature_info.get("feature_layer_info", {}).get("args", {}).get("ops", {})

    if not tf_ops:
        return feature_tensor

    for tf_op in tf_ops:
        try:
            fn_, fn_args = eval(tf_op["fn"]), tf_op.get("args", {})
        except AttributeError as e:
            raise KeyError("Invalid fn specified for tf_native_op : {}\n{}".format(tf_op["fn"], e))
        
        try:
            feature_tensor = fn_(feature_tensor, **fn_args)
        except Exception as e:
            raise Exception("Error while applying {} to {} feature:\n{}".format(tf_op["fn"], feature_node_name, e))

    # Adjusting the shape to the default feature fns for concatenating in the next step
    feature_tensor = tf.expand_dims(feature_tensor, axis=-1, name="{}_tf_native_op".format(feature_node_name))

    return feature_tensor
